"""Artistools - NLTE population related functions."""

import math
import multiprocessing
import re
import string
from functools import lru_cache
from functools import partial
from pathlib import Path

import pandas as pd
from astropy import constants as const

import artistools as at


def texifyterm(strterm: str) -> str:
    """Replace a term string with TeX notation equivalent."""
    strtermtex = ""
    passed_term_Lchar = False

    for termpiece in re.split("([_A-Za-z])", strterm):
        if re.match("[0-9]", termpiece) is not None and not passed_term_Lchar:
            # 2S + 1 number
            strtermtex += r"$^{" + termpiece + r"}$"
        elif re.match("[A-Z]", termpiece) is not None:
            # L character - SPDFGH...
            strtermtex += termpiece
            passed_term_Lchar = True
        elif re.match("[eo]", termpiece) is not None and passed_term_Lchar:
            # odd flag, but don't want to confuse it with the energy index (e.g. o4Fo[2])
            if termpiece != "e":  # even is assumed by default (and looks neater with all the 'e's)
                strtermtex += r"$^{\rm " + termpiece + r"}$"
        elif re.match(r"[0-9]?.*\]", termpiece) is not None:
            # J value
            strtermtex += termpiece.split("[")[0] + r"$_{" + termpiece.lstrip(string.digits).strip("[]") + r"}$"
        elif re.match("[0-9]", termpiece) is not None and passed_term_Lchar:
            # extra number after S char
            strtermtex += termpiece

    return strtermtex.replace("$$", "")


def texifyconfiguration(levelname: str) -> str:
    """Replace a level configuration with the formatted LaTeX equivalent."""
    # the underscore gets confused with LaTeX subscript operator, so switch it to the hash symbol
    strout = "#".join(levelname.split("_")[:-1]) + "#"
    for strorbitalocc in re.findall(r"[0-9][a-z][0-9]?[#(]", strout):
        n, lchar, occ = re.split("([a-z])", strorbitalocc)
        lastchar = "(" if occ.endswith("(") else "#"
        occ = occ.rstrip("#(")
        strorbitalocctex = n + lchar + (r"$^{" + occ + r"}$" if occ else "") + lastchar
        strout = strout.replace(strorbitalocc, strorbitalocctex)

    for parentterm in re.findall(r"\([0-9][A-Z][^)]?\)", strout):
        parentermtex = f'({texifyterm(parentterm.strip("()"))})'
        strout = strout.replace(parentterm, parentermtex)
    strterm = levelname.split("_")[-1]
    strout += " " + texifyterm(strterm)

    return strout.replace("#", "").replace("$$", "")


def add_lte_pops(modelpath, dfpop, columntemperature_tuples, noprint=False, maxlevel=-1):
    """Add columns to dfpop with LTE populations.

    columntemperature_tuples is a sequence of tuples of column name and temperature, e.g., ('mycolumn', 3000)
    """
    k_b = const.k_B.to("eV / K").value

    for _, row in dfpop.drop_duplicates(["modelgridindex", "timestep", "Z", "ion_stage"]).iterrows():
        modelgridindex = int(row.modelgridindex)
        timestep = int(row.timestep)
        Z = int(row.Z)
        ion_stage = int(row.ion_stage)

        ionlevels = at.atomic.get_levels(modelpath).query("Z == @Z and ion_stage == @ion_stage").iloc[0].levels

        gs_g = ionlevels.iloc[0].g
        gs_energy = ionlevels.iloc[0].energy_ev

        # gs_pop = dfpop.query(
        #     "modelgridindex == @modelgridindex and timestep == @timestep "
        #     "and Z == @Z and ion_stage == @ion_stage and level == 0"
        # ).iloc[0]["n_NLTE"]

        masksuperlevel = (
            (dfpop["modelgridindex"] == modelgridindex)
            & (dfpop["timestep"] == timestep)
            & (dfpop["Z"] == Z)
            & (dfpop["ion_stage"] == ion_stage)
            & (dfpop["level"] == -1)
        )

        masknotsuperlevel = (
            (dfpop["modelgridindex"] == modelgridindex)
            & (dfpop["timestep"] == timestep)
            & (dfpop["Z"] == Z)
            & (dfpop["ion_stage"] == ion_stage)
            & (dfpop["level"] != -1)
        )

        def f_ltepop(x, T_exc: float, gsg: float, gse: float, ionlevels) -> float:
            return (
                ionlevels.iloc[int(x.level)].g
                / gsg
                * math.exp(-(ionlevels.iloc[int(x.level)].energy_ev - gse) / k_b / T_exc)
            )

        for columnname, T_exc in columntemperature_tuples:
            dfpop.loc[masknotsuperlevel, columnname] = dfpop.loc[masknotsuperlevel].apply(
                f_ltepop, args=(T_exc, gs_g, gs_energy, ionlevels), axis=1
            )

        if not dfpop[masksuperlevel].empty:
            levelnumber_sl = (
                dfpop.query(
                    "modelgridindex == @modelgridindex and timestep == @timestep "
                    "and Z == @Z and ion_stage == @ion_stage"
                ).level.max()
                + 1
            )

            if maxlevel < 0 or levelnumber_sl <= maxlevel:
                if not noprint:
                    print(
                        f"{at.get_elsymbol(Z)} {at.roman_numerals[ion_stage]} "
                        f"has a superlevel at level {levelnumber_sl}"
                    )

                for columnname, T_exc in columntemperature_tuples:
                    dfpop.loc[masksuperlevel, columnname] = (
                        ionlevels.iloc[levelnumber_sl:]
                        .eval("g / @gs_g * exp(- (energy_ev - @gs_energy) / @k_b / @T_exc)")
                        .sum()
                    )

            dfpop.loc[masksuperlevel, "level"] = levelnumber_sl + 2

    return dfpop


def read_file(nltefilepath: str | Path) -> pd.DataFrame:
    """Read NLTE populations from one file."""
    if not Path(nltefilepath).is_file():
        nltefilepathgz = Path(f"{nltefilepath!s}.gz")
        nltefilepathxz = Path(f"{nltefilepath!s}.xz")
        if nltefilepathxz.is_file():
            nltefilepath = nltefilepathxz
        elif nltefilepathgz.is_file():
            nltefilepath = nltefilepathgz
        else:
            # print(f'Warning: Could not find {nltefilepath}')
            return pd.DataFrame()

    filesize = Path(nltefilepath).stat().st_size / 1024 / 1024
    print(f"Reading {nltefilepath} ({filesize:.2f} MiB)")

    try:
        dfpop = pd.read_csv(nltefilepath, sep=r"\s+")
    except pd.errors.EmptyDataError:
        return pd.DataFrame()
    if "ion_stage" in dfpop.columns:
        dfpop = dfpop.rename(columns={"ion_stage": "ion_stage"})

    return dfpop


def read_file_filtered(nltefilepath, strquery=None, dfqueryvars=None):
    dfpopfile = read_file(nltefilepath)

    if strquery and not dfpopfile.empty:
        dfpopfile = dfpopfile.query(strquery, local_dict=dfqueryvars)

    return dfpopfile


@lru_cache(maxsize=2)
def read_files(
    modelpath: str | Path,
    timestep: int = -1,
    modelgridindex: int = -1,
    dfquery: str | None = None,
    dfqueryvars: dict | None = None,
) -> pd.DataFrame:
    """Read in NLTE populations from a model for a particular timestep and grid cell."""
    if dfqueryvars is None:
        dfqueryvars = {}

    mpiranklist = at.get_mpiranklist(modelpath, modelgridindex=modelgridindex)

    nltefilepaths = [
        Path(folderpath, f"nlte_{mpirank:04d}.out")
        for folderpath in at.get_runfolders(modelpath, timestep=timestep)
        for mpirank in mpiranklist
    ]

    dfqueryvars["modelgridindex"] = modelgridindex
    dfqueryvars["timestep"] = timestep

    dfquery_full = "timestep==@timestep" if timestep >= 0 else ""
    if modelgridindex >= 0:
        if dfquery_full:
            dfquery_full += " and "
        dfquery_full += "modelgridindex==@modelgridindex"

    if dfquery:
        if dfquery_full:
            dfquery_full = f"({dfquery_full}) and "
        dfquery_full += f"({dfquery})"

    if at.get_config()["num_processes"] > 1:
        with multiprocessing.Pool(processes=at.get_config()["num_processes"]) as pool:
            arr_dfnltepop = pool.map(
                partial(read_file_filtered, strquery=dfquery_full, dfqueryvars=dfqueryvars), nltefilepaths
            )
            pool.close()
            pool.join()
            pool.terminate()
    else:
        arr_dfnltepop = [read_file_filtered(f, strquery=dfquery_full, dfqueryvars=dfqueryvars) for f in nltefilepaths]

    return pd.concat(arr_dfnltepop).copy()



"""Artistools - spectra related functions."""

from artistools.nonthermal import solvespencerfanocmd
from artistools.nonthermal._nonthermal_core import analyse_ntspectrum
from artistools.nonthermal._nonthermal_core import ar_xs
from artistools.nonthermal._nonthermal_core import calculate_frac_heating
from artistools.nonthermal._nonthermal_core import calculate_Latom_excitation
from artistools.nonthermal._nonthermal_core import calculate_Latom_ionisation
from artistools.nonthermal._nonthermal_core import calculate_N_e
from artistools.nonthermal._nonthermal_core import calculate_nt_frac_excitation
from artistools.nonthermal._nonthermal_core import differentialsfmatrix_add_ionization_shell
from artistools.nonthermal._nonthermal_core import e_s_test
from artistools.nonthermal._nonthermal_core import get_arxs_array_ion
from artistools.nonthermal._nonthermal_core import get_arxs_array_shell
from artistools.nonthermal._nonthermal_core import get_electronoccupancy
from artistools.nonthermal._nonthermal_core import get_energyindex_gteq
from artistools.nonthermal._nonthermal_core import get_energyindex_lteq
from artistools.nonthermal._nonthermal_core import get_epsilon_avg
from artistools.nonthermal._nonthermal_core import get_fij_ln_en_ionisation
from artistools.nonthermal._nonthermal_core import get_J
from artistools.nonthermal._nonthermal_core import get_Latom_axelrod
from artistools.nonthermal._nonthermal_core import get_Lelec_axelrod
from artistools.nonthermal._nonthermal_core import get_lotz_xs_ionisation
from artistools.nonthermal._nonthermal_core import get_mean_binding_energy
from artistools.nonthermal._nonthermal_core import get_mean_binding_energy_alt
from artistools.nonthermal._nonthermal_core import get_nne
from artistools.nonthermal._nonthermal_core import get_nne_nt
from artistools.nonthermal._nonthermal_core import get_nnetot
from artistools.nonthermal._nonthermal_core import get_nntot
from artistools.nonthermal._nonthermal_core import get_xs_excitation
from artistools.nonthermal._nonthermal_core import get_xs_excitation_vector
from artistools.nonthermal._nonthermal_core import get_Zbar
from artistools.nonthermal._nonthermal_core import get_Zboundbar
from artistools.nonthermal._nonthermal_core import lossfunction
from artistools.nonthermal._nonthermal_core import lossfunction_axelrod
from artistools.nonthermal._nonthermal_core import namedtuple
from artistools.nonthermal._nonthermal_core import Psecondary
from artistools.nonthermal._nonthermal_core import read_binding_energies
from artistools.nonthermal._nonthermal_core import read_colliondata
from artistools.nonthermal._nonthermal_core import sfmatrix_add_excitation
from artistools.nonthermal._nonthermal_core import sfmatrix_add_ionization_shell
from artistools.nonthermal._nonthermal_core import solve_spencerfano_differentialform
from artistools.nonthermal._nonthermal_core import workfunction_tests
from artistools.nonthermal.plotnonthermal import addargs
from artistools.nonthermal.plotnonthermal import main as plot



#!/usr/bin/env python3
import argparse
import math
import typing as t
from pathlib import Path

import numpy as np
import pandas as pd

import artistools as at


def addargs(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("-outputpath", "-o", default=".", help="Path for output files")


def main(args: argparse.Namespace | None = None, argsraw: t.Sequence[str] | None = None, **kwargs) -> None:
    """Create solar r-process pattern in ARTIS format."""
    if args is None:
        parser = argparse.ArgumentParser(formatter_class=at.CustomArgHelpFormatter, description=__doc__)

        addargs(parser)
        at.set_args_from_dict(parser, kwargs)
        args = parser.parse_args([] if kwargs else argsraw)

    dfsolarabund = pd.read_csv(
        at.get_config()["path_datadir"] / "solar_r_abundance_pattern.txt", sep=r"\s+", comment="#"
    )

    dfsolarabund["radioactive"] = True

    # print(dfsolarabund)

    dfbetaminus = pd.read_csv(
        at.get_config()["path_datadir"] / "betaminusdecays.txt",
        sep=r"\s+",
        comment="#",
        names=["A", "Z", "Q[MeV]", "Egamma[MeV]", "Eelec[MeV]", "Eneutrino[MeV]", "tau[s]"],
    )

    def undecayed_z(row):
        dfmasschain = dfbetaminus.query("A == @row.A", inplace=False)
        return int(row.Z) if dfmasschain.empty else int(dfmasschain.Z.min())

    dfsolarabund_undecayed = dfsolarabund.copy()
    dfsolarabund_undecayed["Z"] = dfsolarabund_undecayed.apply(undecayed_z, axis=1)

    # Andreas uses 90% Fe and the rest solar
    dfsolarabund_undecayed = (
        dfsolarabund_undecayed.append({"Z": 26, "A": 56, "numberfrac": 0.005, "radioactive": False}, ignore_index=True)
        .append({"Z": 27, "A": 59, "numberfrac": 0.005, "radioactive": False}, ignore_index=True)
        .append({"Z": 28, "A": 58, "numberfrac": 0.005, "radioactive": False}, ignore_index=True)
    )

    normfactor = (  # noqa: F841
        dfsolarabund_undecayed.numberfrac.sum()
    )  # convert number fractions in solar to fractions of r-process
    dfsolarabund_undecayed = dfsolarabund_undecayed.eval("numberfrac = numberfrac / @normfactor").eval(
        "massfrac = numberfrac * A"
    )
    massfracnormfactor = dfsolarabund_undecayed.massfrac.sum()  # noqa: F841
    dfsolarabund_undecayed = dfsolarabund_undecayed.eval("massfrac = massfrac / @massfracnormfactor")

    # print(dfsolarabund_undecayed)

    t_model_init_days = 0.000231481
    t_model_init_seconds = t_model_init_days * 24 * 60 * 60  # noqa: F841

    wollager_profilename = "wollager_ejectaprofile_10bins.txt"
    if Path(wollager_profilename).exists():
        with Path(wollager_profilename).open("rt") as f:
            t_model_init_days_in = float(f.readline().strip().removesuffix(" day"))
        dfdensities = pd.read_csv(
            wollager_profilename, sep=r"\s+", skiprows=1, names=["cellid", "vel_r_max_kmps", "rho"]
        )
        dfdensities["cellid"] = dfdensities["cellid"].astype(int)
        dfdensities["vel_r_min_kmps"] = np.concatenate(([0.0], dfdensities["vel_r_max_kmps"].to_numpy()[:-1]))

        t_model_init_seconds_in = t_model_init_days_in * 24 * 60 * 60  # noqa: F841
        dfdensities = dfdensities.eval(
            "mass_g = rho * 4. / 3. * @math.pi * (vel_r_max_kmps ** 3 - vel_r_min_kmps ** 3)"
            "* (1e5 * @t_model_init_seconds_in) ** 3",
        )

        # now replace the density at the input time with the density at required time

        dfdensities = dfdensities.eval(
            "rho = mass_g / ("
            "4. / 3. * @math.pi * (vel_r_max_kmps ** 3 - vel_r_min_kmps ** 3)"
            " * (1e5 * @t_model_init_seconds) ** 3)",
        )
    else:
        dfdensities = pd.DataFrame({"rho": 10**-3, "vel_r_max_kmps": 6.0e4}, index=[0])

    # print(dfdensities)
    cellcount = len(dfdensities)
    # write abundances.txt

    dictelemabund = {}
    for atomic_number in range(1, dfsolarabund_undecayed.Z.max() + 1):
        dictelemabund[f"X_{at.get_elsymbol(atomic_number)}"] = dfsolarabund_undecayed.query(
            "Z == @atomic_number", inplace=False
        ).massfrac.sum()

    dfelabundances = pd.DataFrame([{"inputcellid": mgi + 1} | dictelemabund for mgi in range(cellcount)])
    # print(dfelabundances)
    at.inputmodel.save_initelemabundances(dfelabundances=dfelabundances, outpath=Path(args.outputpath))

    # write model.txt

    rowdict = {
        # 'inputcellid': 1,
        # 'vel_r_max_kmps': 6.e4,
        # 'logrho': -3.,
        "X_Fegroup": 1.0,
        "X_Ni56": 0.0,
        "X_Co56": 0.0,
        "X_Fe52": 0.0,
        "X_Cr48": 0.0,
        "X_Ni57": 0.0,
        "X_Co57": 0.0,
    }

    for _, row in dfsolarabund_undecayed.query("radioactive == True").iterrows():
        rowdict[f"X_{at.get_elsymbol(int(row.Z))}{int(row.A)}"] = row.massfrac

    modeldata = [
        (
            {
                "inputcellid": mgi + 1,
                "vel_r_max_kmps": densityrow["vel_r_max_kmps"],
                "logrho": math.log10(densityrow["rho"]),
            }
            | rowdict
        )
        for mgi, densityrow in dfdensities.iterrows()
    ]

    dfmodel = pd.DataFrame(modeldata)
    # print(dfmodel)
    at.inputmodel.save_modeldata(dfmodel=dfmodel, t_model_init_days=t_model_init_days, outpath=args.outputpath)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
import argparse
import typing as t
from pathlib import Path

import argcomplete
import matplotlib.pyplot as plt
import polars as pl

import artistools as at


def addargs(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "modelpath",
        default=[],
        nargs="*",
        action=at.AppendPath,
        help="Path(s) to model.txt file(s) or folders containing model.txt)",
    )
    parser.add_argument("-outputpath", "-o", default=".", help="Path for output files")


def main(args: argparse.Namespace | None = None, argsraw: t.Sequence[str] | None = None, **kwargs) -> None:
    """Plot the radial density profile of an ARTIS model."""
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description=__doc__,
        )

        addargs(parser)
        at.set_args_from_dict(parser, kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args([] if kwargs else argsraw)

    fig, axes = plt.subplots(
        nrows=2,
        ncols=1,
        sharex=True,
        sharey=False,
        figsize=(8, 8),
        tight_layout={"pad": 0.4, "w_pad": 0.0, "h_pad": 0.0},
    )

    if not args.modelpath:
        args.modelpath = ["."]

    for modelpath in args.modelpath:
        dfmodel, modelmeta = at.get_modeldata_polars(
            modelpath, derived_cols=["vel_r_min", "vel_r_mid", "vel_r_max", "mass_g"]
        )
        label = at.get_model_name(modelpath)
        binned_xvals: list[float] = []
        binned_yvals: list[float] = []

        # total_mass = dfmodel.mass_g.sum() / 1.989e33
        dfmodel = dfmodel.sort(by="vel_r_mid")

        dfmodelcollect = dfmodel.select(["modelgridindex", "vel_r_min", "vel_r_mid", "vel_r_max", "mass_g"]).collect()

        vuppers = dfmodelcollect["vel_r_max"].unique().sort()
        enclosed_xvals = [0.0, *(vuppers / 29979245800).to_list()]
        enclosed_yvals = [0.0] + [
            dfmodelcollect.filter(pl.col("vel_r_mid") <= vupper)["mass_g"].sum() / 1.989e33 for vupper in vuppers
        ]
        axes[0].plot(enclosed_xvals, enclosed_yvals, label=label)

        if "vel_r_max_kmps" in dfmodel.columns:
            vupperscoarse = vuppers.to_list()
        else:
            ncoarsevelbins = int(
                modelmeta["ncoordgridrcyl"] if "ncoordgridrcyl" in modelmeta else modelmeta["ncoordgridx"] / 2.0
            )
            vupperscoarse = [modelmeta["vmax_cmps"] / ncoarsevelbins * (i + 1) for i in range(ncoarsevelbins)]

        vlowerscoarse = [0.0] + vupperscoarse[:-1]
        for vlower, vupper in zip(vlowerscoarse, vupperscoarse):
            velbinmass = (
                dfmodelcollect.filter(pl.col("vel_r_mid").is_between(vlower, vupper, closed="left"))["mass_g"].sum()
                / 1.989e33
            )
            assert vlower < vupper
            binned_xvals.extend((vlower / 29979245800, vupper / 29979245800))
            delta_beta = (vupper - vlower) / 29979245800
            yval = velbinmass / delta_beta
            binned_yvals.extend((yval, yval))

        axes[1].plot(binned_xvals, binned_yvals, label=label)
        vmax_on_c = modelmeta["vmax_cmps"] / 29979245800
        axes[0].set_xlim(right=vmax_on_c)

    axes[-1].set_xlabel("Velocity [$c$]")
    axes[0].set_ylabel(r"Mass Enclosed [M$_\odot$]")
    axes[1].set_ylabel(r"$\Delta$M/$\Delta v$  [M$_\odot/c$]")
    axes[0].legend()

    axes[-1].set_xlim(left=0.0)
    axes[0].set_ylim(bottom=0.0)
    axes[1].set_ylim(bottom=0.0)

    outfilepath = Path(args.outputpath)
    if outfilepath.is_dir():
        outfilepath /= "densityprofile.pdf"

    plt.savefig(outfilepath)
    print(f"Saved {outfilepath}")


if __name__ == "__main__":
    main()



"""Artistools - non-LTE population functions."""

from artistools.nltepops.nltepops import add_lte_pops
from artistools.nltepops.nltepops import read_file
from artistools.nltepops.nltepops import read_file_filtered
from artistools.nltepops.nltepops import read_files
from artistools.nltepops.nltepops import texifyconfiguration
from artistools.nltepops.nltepops import texifyterm
from artistools.nltepops.plotnltepops import addargs
from artistools.nltepops.plotnltepops import main as plot



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
import argparse
import string
import typing as t
from pathlib import Path

import argcomplete
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import artistools as at


def get_model_recombenergy(dfbinding, args):
    args.inputfile = Path(
        "/Users/luke/Library/Mobile"
        " Documents/com~apple~CloudDocs/artis_runs/kilonova_SFHo_1d_0p05day_ufix_betaminusthermdelay_fixinitialye_verylate_redo"
    )

    print(f"Reading {args.inputfile}")
    dfmodel, t_model_init_days, vmax = at.inputmodel.get_modeldata_tuple(args.inputfile, get_elemabundances=False)

    t_model_init_seconds = t_model_init_days * 24 * 60 * 60
    print(f"Model is defined at {t_model_init_days} days ({t_model_init_seconds:.4f} seconds)")

    msun_g = 1.989e33
    amu_g = 1.66054e-24  # 1 atomic mass unit in grams
    ev_erg = 1.60218e-12  # 1 eV in erg
    mass_msun_rho = dfmodel["mass_g"].sum() / msun_g

    mass_msun_accounted = 0.0
    model_el_binding_en_ev = {}
    model_tot_binding_en_ev = 0.0
    for column in dfmodel.columns:
        if column.startswith("X_"):
            species = column.replace("X_", "")
            speciesabund_g = np.dot(dfmodel[column], dfmodel["mass_g"])

            species_mass_msun = speciesabund_g / msun_g

            if species[-1].isdigit():  # isotopic species
                atomic_number = at.get_atomic_number(species)
                elsymb = at.get_elsymbol(atomic_number)
                massnumber = int(species.lstrip(string.ascii_letters))
                matchrows = dfbinding.query("Z == @atomic_number")

                binding_en_ev = 0.0 if matchrows.empty else matchrows.iloc[0]["TotBEn"]

                # print(species, atomic_number, massnumber, el_binding_en_ev)
                contrib_binding_en_ev = speciesabund_g / (massnumber * amu_g) * binding_en_ev

                model_el_binding_en_ev[elsymb] = model_el_binding_en_ev.get(elsymb, 0.0) + contrib_binding_en_ev
                mass_msun_accounted += species_mass_msun
                model_tot_binding_en_ev += contrib_binding_en_ev

            elif species.lower() != "fegroup":  # ignore special group abundance
                pass

    def sortkey(item):
        return at.get_atomic_number(item[0])

    for elsymb, binding_en_ev in sorted(model_el_binding_en_ev.items(), key=sortkey):
        zstr = f"Z={atomic_number}"
        print(f"{zstr:>5} {binding_en_ev * ev_erg} erg")

    print(f"Total electron binding energy {model_tot_binding_en_ev * ev_erg:.3e} erg")
    print(f"Mass from density {mass_msun_rho:.3e}")
    print(f"Mass from sum of isotopes {mass_msun_accounted:.3e}")


def get_particle_elec_binding_energy_per_gram(traj_root, dictbinding, particleid, time_s):
    # find the closest timestep to the required time
    nts = at.inputmodel.rprocess_from_trajectory.get_closest_network_timestep(traj_root, particleid, time_s)
    memberfilename = f"./Run_rprocess/nz-plane{nts:05d}"

    dftrajnucabund, traj_time_s = at.inputmodel.rprocess_from_trajectory.get_trajectory_timestepfile_nuc_abund(
        traj_root=traj_root, particleid=particleid, memberfilename=memberfilename
    )

    dftrajnucabund["Z_be_tot_ev"] = [dictbinding.get(Z, 0.0) for Z in dftrajnucabund["Z"]]

    amu_g = 1.66054e-24  # 1 atomic mass unit in grams  # noqa: F841

    # frac_unaccounted = dftrajnucabund[dftrajnucabund['Z_be_tot_ev'] == 0].massfrac.sum()
    # print(f'frac_unaccounted {frac_unaccounted}')
    # if frac_unaccounted > .3:
    #     print(dftrajnucabund)
    # assert frac_unaccounted < 0.3

    dftrajnucabund = dftrajnucabund.eval("recombenergy_ev_per_gram = Z_be_tot_ev * massfrac / (Z + N) / @amu_g")

    # contrib_binding_en_ev = speciesabund_g / (massnumber * amu_g) * binding_en_ev

    # print(dftrajnucabund)

    return dftrajnucabund["recombenergy_ev_per_gram"].sum()


def get_particle_nucenergy_released(traj_root, particleid, tmin_s, time_s_end):
    memberfilename = "./Run_rprocess/energy_thermo.dat"
    erg_to_ev = 6.242e11
    with at.inputmodel.rprocess_from_trajectory.open_tar_file_or_extracted(
        traj_root=traj_root, particleid=particleid, memberfilename=memberfilename
    ) as fthermo:
        dfthermo = pd.read_csv(fthermo, sep=r"\s+", usecols=["#count", "time/s", "Qdot", "Ye"])
        dfthermo = dfthermo.rename(columns={"time/s": "time_s"})
        dfthermo = dfthermo.query("time_s >= @tmin_s")
        dfthermo = dfthermo.query("time_s <= @time_s_end")
        return np.trapz(y=dfthermo["Qdot"], x=dfthermo["time_s"]) * erg_to_ev
        # print(dfthermo)


def get_particles_recomb_nuc_energy(traj_root, dfbinding):
    # sourcery skip: use-contextlib-suppress
    dfsnapshot = at.inputmodel.modelfromhydro.read_ejectasnapshot(
        "/Users/luke/Library/Mobile Documents/com~apple~CloudDocs/Archive/Astronomy/Mergers/SFHo_snapshot"
    ).sort_values("ye")

    dictbinding = dict(dfbinding[["Z", "TotBEn"]].itertuples(index=False))

    tmin_s = 10
    time_s = 6 * 3600

    ye_list = []
    elecbinding_en_list = []
    nuclear_released_en_list = []
    for particleid, ye, _pmass in dfsnapshot[["id", "ye", "pmass"]].itertuples(index=False):
        try:
            elecbinding_en = get_particle_elec_binding_energy_per_gram(
                traj_root=traj_root, dictbinding=dictbinding, particleid=particleid, time_s=time_s
            )
            nuc_en_released = get_particle_nucenergy_released(
                traj_root=traj_root, particleid=particleid, tmin_s=tmin_s, time_s_end=time_s
            )
            ye_list.append(ye)
            elecbinding_en_list.append(elecbinding_en)
            nuclear_released_en_list.append(nuc_en_released)
            # print(particleid, ye, elecbinding_en)
        except FileNotFoundError:
            pass
            # print(f' WARNING particle {particleid} not found! ')

    dfrecomb = pd.DataFrame(
        {
            "ye": ye_list,
            "recombenergy_ev_per_gram": elecbinding_en_list,
            "nuclear_released_ev_per_gram": nuclear_released_en_list,
        }
    )

    print(dfrecomb)

    fig, ax = plt.subplots(
        nrows=1,
        ncols=1,
        sharex=True,
        sharey=False,
        figsize=(6, 4),
        tight_layout={"pad": 0.4, "w_pad": 0.0, "h_pad": 0.0},
    )

    ax.plot(
        dfrecomb["ye"],
        dfrecomb["recombenergy_ev_per_gram"],
        label=f"electron binding for composition at {time_s:.1e} s [eV/g]",
        lw=0,
        marker=".",
    )
    ax.plot(
        dfrecomb["ye"],
        dfrecomb["nuclear_released_ev_per_gram"],
        label=f"nuclear release between {tmin_s:.1e} s and {time_s:.1e} s [eV/g]",
        lw=0,
        marker=".",
    )
    ax.legend(loc="best", handlelength=2, frameon=False, numpoints=1)
    plt.xlabel("Ye")
    # plt.ylabel('eV / g')
    plt.yscale("log")
    ax.set_ylim(bottom=1e24, top=1e33)

    plt.savefig("recomb.pdf", format="pdf")


def addargs(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-inputfile", "-i", default="model.txt", help="Path of input file or folder containing model.txt"
    )


def main(args: argparse.Namespace | None = None, argsraw: t.Sequence[str] | None = None, **kwargs: t.Any) -> None:
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description="Get the recombination energy from fully ionised to fully neutral for an input model.",
        )

        addargs(parser)
        at.set_args_from_dict(parser, kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args([] if kwargs else argsraw)

    with Path(at.get_config()["path_datadir"], "ElBiEn_2007.txt").open() as fbinding:
        for _ in range(11):
            header = fbinding.readline().lstrip(" #").split()
        # print(header)
        dfbinding = pd.read_csv(fbinding, sep=r"\s+", names=header)
        # print(dfbinding)

    traj_root = Path(
        Path.home() / "Google Drive/Shared Drives/GSI NSM/Mergers/SFHo_long/Trajectory_SFHo_long-radius-entropy"
    )
    # get_model_recombenergy(dfbinding, args)
    get_particles_recomb_nuc_energy(traj_root, dfbinding)


if __name__ == "__main__":
    main()



import os
import tarfile
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import integrate

import artistools as at

DAY = 86400  # day in seconds
MSUN = 1.989e33  # solar mass in grams


def write_energydistribution_file(energydistdata, outputfilepath="."):
    print("Writing energydistribution.txt")
    with Path(outputfilepath, "energydistribution.txt").open("w") as fmodel:
        fmodel.write(f'{len(energydistdata["cell_energy"])}\n')  # write number of points
        energydistdata.to_csv(fmodel, header=False, sep="\t", index=False, float_format="%g")


def write_energyrate_file(energy_rate_data, outputfilepath="."):
    print("Writing energyrate.txt")
    with Path(outputfilepath, "energyrate.txt").open("w") as fmodel:
        fmodel.write(f'{len(energy_rate_data["times"])}\n')  # write number of points
        energy_rate_data.to_csv(fmodel, sep="\t", index=False, header=False, float_format="%.10f")


def rprocess_const_and_powerlaw():
    """Following eqn 4 Korobkin 2012."""

    def integrand(t_days, t0, epsilon0, sigma, alpha, thermalisation_factor):
        return (epsilon0 * ((1 / 2) - (1 / np.pi * np.arctan((t_days - t0) / sigma))) ** alpha) * (
            thermalisation_factor / 0.5
        )

    from scipy.integrate import quad

    tmin = 0.01 * DAY
    tmax = 50 * DAY
    t0 = 1.3  # seconds
    epsilon0 = 2e18
    sigma = 0.11
    alpha = 1.3
    thermalisation_factor = 0.5

    E_tot = quad(integrand, tmin, tmax, args=(t0, epsilon0, sigma, alpha, thermalisation_factor))  # ergs/s/g
    print("Etot per gram", E_tot[0])
    E_tot = E_tot[0]

    times = np.logspace(np.log10(tmin), np.log10(tmax), num=200)
    energy_per_gram_cumulative = [0]
    for time in times[1:]:
        cumulative_integral = quad(
            integrand, tmin, time, args=(t0, epsilon0, sigma, alpha, thermalisation_factor)
        )  # ergs/s/g
        energy_per_gram_cumulative.append(cumulative_integral[0])

    energy_per_gram_cumulative = np.array(energy_per_gram_cumulative)

    rate = energy_per_gram_cumulative / E_tot

    # nuclear_heating_power = [integrand(time, t0, epsilon0, sigma, alpha, thermalisation_factor) for time in times]

    # times_and_rate = {'times': times/DAY, 'rate': rate, 'nuclear_heating_power': nuclear_heating_power}
    times_and_rate = {"times": times / DAY, "rate": rate}
    times_and_rate = pd.DataFrame(data=times_and_rate)

    return times_and_rate, E_tot


def define_heating_rate():
    tmin = 0.0001  # days
    tmax = 50

    times = np.logspace(np.log10(tmin), np.log10(tmax), num=300)  # days
    qdot = 5e9 * (times) ** (-1.3)  # define energy power law (5e9*t^-1.3)

    E_tot = integrate.trapezoid(y=qdot, x=times)  # ergs/s/g
    # print("Etot per gram", E_tot, E_tot*1.989e33*0.01)

    cumulative_integrated_energy = integrate.cumulative_trapezoid(y=qdot, x=times)
    cumulative_integrated_energy = np.insert(cumulative_integrated_energy, 0, 0)

    rate = cumulative_integrated_energy / E_tot

    times_and_rate = {"times": times, "rate": rate}
    times_and_rate = pd.DataFrame(data=times_and_rate)

    dE = np.diff(times_and_rate["rate"] * E_tot)
    dt = np.diff(times * 24 * 60 * 60)

    intergrated_rate = dE / dt
    scale_factor_energy_diff = max(qdot[1:] / intergrated_rate)
    print(np.mean(scale_factor_energy_diff))
    E_tot *= scale_factor_energy_diff
    # print(f"E_tot after integrated line scaled to match energy of power law: {E_tot}")

    dE = np.diff(times_and_rate["rate"] * E_tot)
    dt = np.diff(times * 24 * 60 * 60)

    # check energy rate is on top of power law line
    # plt.plot(times_and_rate["times"][1:], (dE / dt) * 0.01 * MSUN)
    # plt.plot(times_and_rate["times"], qdot * 0.01 * MSUN)
    # plt.yscale("log")
    # plt.xscale("log")

    # plt.xlabel("Time [days]")
    # plt.ylabel("Q [erg/g/s]")
    # # plt.xlim(0.1, 20)
    # # plt.ylim(5e39, 2e41)
    # plt.show()

    return times_and_rate, E_tot


def energy_from_rprocess_calculation(energy_thermo_data, get_rate=True):
    index_time_greaterthan = energy_thermo_data[energy_thermo_data["time/s"] > 1e7].index  # 1e7 seconds = 116 days
    energy_thermo_data = energy_thermo_data.drop(index_time_greaterthan)
    # print("Dropping times later than 116 days")

    skipfirstnrows = 0  # not sure first values look sensible -- check this
    times = energy_thermo_data["time/s"][skipfirstnrows:]
    qdot = energy_thermo_data["Qdot"][skipfirstnrows:]

    E_tot = np.trapz(y=qdot, x=times)  # erg / g

    if get_rate:
        print(f"E_tot {E_tot} erg/g")

        cumulative_integrated_energy = integrate.cumulative_trapezoid(y=qdot, x=times)
        cumulative_integrated_energy = np.insert(cumulative_integrated_energy, 0, 0)

        rate = cumulative_integrated_energy / E_tot

        times_and_rate = {"times": times / DAY, "rate": rate}
        times_and_rate = pd.DataFrame(data=times_and_rate)

        return times_and_rate, E_tot

    return E_tot


def get_rprocess_calculation_files(path_to_rprocess_calculation, interpolate_trajectories=False, thermalisation=False):
    tarfiles = [file for file in os.listdir(path_to_rprocess_calculation) if file.endswith(".tar.xz")]

    trajectory_ids = []
    trajectory_E_tot = []

    if interpolate_trajectories:
        interpolated_trajectories = {"time/s": np.logspace(-1, 7, 300)}

    energy_thermo_filepath = "./Run_rprocess/energy_thermo.dat"
    for file in tarfiles:
        trajectory_id = file.split(".")[0]
        tar = tarfile.open(path_to_rprocess_calculation / file, mode="r:*")

        energythermo_file = tar.extractfile(member=energy_thermo_filepath)
        energy_thermo_data = pd.read_csv(energythermo_file, sep=r"\s+")
        # print(energy_thermo_data['Qdot'])
        # print(energy_thermo_data['time/s'])

        if interpolate_trajectories:
            qdotinterp = np.interp(
                interpolated_trajectories["time/s"], energy_thermo_data["time/s"], energy_thermo_data["Qdot"]
            )
            interpolated_trajectories[trajectory_id] = qdotinterp

        E_tot = energy_from_rprocess_calculation(energy_thermo_data, get_rate=False)

        trajectory_ids.append(float(trajectory_id))
        trajectory_E_tot.append(E_tot)

    if interpolate_trajectories:
        interpolated_trajectories = pd.DataFrame.from_dict(interpolated_trajectories)
        interpolated_trajectories["mean"] = interpolated_trajectories.iloc[:, 1:].mean(axis=1)

        index_time_lessthan = interpolated_trajectories[interpolated_trajectories["time/s"] < 1.1e-1].index
        interpolated_trajectories = interpolated_trajectories.drop(index_time_lessthan)

        interpolated_trajectories.to_csv(path_to_rprocess_calculation / "interpolatedQdot.dat", sep=" ", index=False)
    print(f"sum etot {sum(trajectory_E_tot)}")
    trajectory_energy = pd.DataFrame.from_dict({"id": trajectory_ids, "E_tot": trajectory_E_tot}).sort_values(by="id")

    print(trajectory_energy)
    trajectory_energy.to_csv(path_to_rprocess_calculation / "trajectoryQ.dat", sep=" ", index=False)


def make_energydistribution_weightedbyrho(rho, E_tot_per_gram, Mtot_grams):
    print(f"energy distribution weighted by rho (E_tot per gram {E_tot_per_gram})")
    Etot = E_tot_per_gram * Mtot_grams
    print("Etot", Etot)
    numberofcells = len(rho)

    cellenergy = np.array([Etot] * numberofcells)
    cellenergy *= rho / sum(rho)

    energydistdata = {"cellid": np.arange(1, len(rho) + 1), "cell_energy": cellenergy}

    print(f"sum energy cells {sum(energydistdata['cell_energy'])} should equal Etot")
    return pd.DataFrame(data=energydistdata)


def make_energy_files(rho, Mtot_grams, outputpath=None, modelpath=None, model=None):
    powerlaw = True
    if powerlaw:
        print("Using power law for energy rate")
        # times_and_rate, E_tot_per_gram = rprocess_const_and_powerlaw()
        times_and_rate, E_tot_per_gram = define_heating_rate()
    # else:
    #     path = Path(".")
    #     energy_thermo_data = pd.read_csv(path / "interpolatedQdot.dat", sep='\s+')
    #     energy_thermo_data = energy_thermo_data.rename(columns={"mean": "Qdot"})
    #     print("Taking rate from averaged trajectories Qdot")
    #     with pd.option_context("display.max_rows", None, "display.max_columns", None):
    #         print(energy_thermo_data["time/s"] / DAY)
    #     times_and_rate, E_tot_per_gram = energy_from_rprocess_calculation(
    #         energy_thermo_data, get_rate=True, thermalisation=True
    #     )

    weight_energy_by_rho = True
    if weight_energy_by_rho:
        energydistributiondata = make_energydistribution_weightedbyrho(rho, E_tot_per_gram, Mtot_grams)
    # else:
    # energydistributiondata = energy_distribution_from_Q_rprocess(modelpath, model)

    write_energydistribution_file(energydistributiondata, outputfilepath=outputpath)
    write_energyrate_file(times_and_rate, outputfilepath=outputpath)


def plot_energy_rate(modelpath):
    times_and_rate, E_tot = at.inputmodel.energyinputfiles.rprocess_const_and_powerlaw()
    model, _ = at.inputmodel.get_modeldata(modelpath)
    Mtot_grams = model["mass_g"].sum()
    plt.plot(
        times_and_rate["times"], np.array(times_and_rate["nuclear_heating_power"]) * Mtot_grams, color="k", zorder=10
    )


def get_etot_fromfile(modelpath):
    energydistribution_data = pd.read_csv(
        Path(modelpath) / "energydistribution.txt",
        skiprows=1,
        sep=r"\s+",
        header=None,
        names=["cellid", "cell_energy"],
    )
    etot = energydistribution_data["cell_energy"].sum()
    return etot, energydistribution_data


def get_energy_rate_fromfile(modelpath):
    return pd.read_csv(Path(modelpath) / "energyrate.txt", skiprows=1, sep=r"\s+", header=None, names=["times", "rate"])



"""Write out ARTIS spectra for each timestep to individual text files."""

import argparse
import typing as t
from pathlib import Path

import argcomplete
import pandas as pd

import artistools as at


def write_spectrum(dfspectrum: pd.DataFrame, outfilepath: Path) -> None:
    with outfilepath.open("w") as spec_file:
        spec_file.write("#lambda f_lambda_1Mpc\n")
        spec_file.write("#[A] [erg/s/cm2/A]\n")
        dfspectrum = dfspectrum[(dfspectrum.lambda_angstroms > 1500) & (dfspectrum.lambda_angstroms < 60000)]

        dfspectrum.to_csv(spec_file, header=False, sep=" ", index=False, columns=["lambda_angstroms", "f_lambda"])
    print(f"Saved {outfilepath}")


def write_flambda_spectra(modelpath: Path) -> None:
    """Write out spectra to text files.

    Writes lambda_angstroms and f_lambda to .txt files for all timesteps and create
    a text file containing the time in days for each timestep.
    """
    outdirectory = Path(modelpath, "spectra")

    outdirectory.mkdir(parents=True, exist_ok=True)

    tmids = at.get_timestep_times(modelpath, loc="mid")

    tslast, tmin_d_valid, tmax_d_valid = at.get_escaped_arrivalrange(modelpath)

    timesteps = [ts for ts in range(tslast + 1) if tmids[ts] >= tmin_d_valid and tmids[ts] <= tmax_d_valid]

    for timestep in timesteps:
        dfspectrum = at.spectra.get_spectrum(modelpath=modelpath, timestepmin=timestep, timestepmax=timestep)[
            -1
        ].to_pandas()

        write_spectrum(dfspectrum, outfilepath=outdirectory / f"spectrum_ts{timestep:02.0f}_{tmids[timestep]:.2f}d.txt")

    for timestep in timesteps:
        if dfspectrum_polar := at.spectra.get_spectrum(
            modelpath=modelpath, timestepmin=timestep, timestepmax=timestep, average_over_phi=True, directionbins=[0]
        ).get(0, None):
            write_spectrum(
                dfspectrum_polar,
                outfilepath=outdirectory / f"spectrum_polar00_ts{timestep:02.0f}_{tmids[timestep]:.2f}d.txt",
            )
        else:
            break


def addargs(parser) -> None:
    parser.add_argument(
        "-modelpath",
        type=Path,
        default=Path(),
        help="Path to ARTIS folder",
    )


def main(args: argparse.Namespace | None = None, argsraw: t.Sequence[str] | None = None, **kwargs) -> None:
    """Plot spectra from ARTIS and reference data."""
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description=__doc__,
        )
        addargs(parser)
        at.set_args_from_dict(parser, kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args([] if kwargs else argsraw)

    write_flambda_spectra(args.modelpath)


if __name__ == "__main__":
    main()



"""artistools.

A collection of plotting, analysis, and file format conversion tools
for the ARTIS radiative transfer code.
"""

import typing as t

from artistools import atomic
from artistools import codecomparison
from artistools import commands
from artistools import deposition
from artistools import estimators
from artistools import gsinetwork
from artistools import inputmodel
from artistools import lightcurve
from artistools import macroatom
from artistools import nltepops
from artistools import nonthermal
from artistools import packets
from artistools import plotspherical
from artistools import plottools
from artistools import radfield
from artistools import spectra
from artistools import transitions
from artistools import writecomparisondata
from artistools.__main__ import addargs
from artistools.__main__ import main
from artistools.configuration import get_config
from artistools.configuration import set_config
from artistools.estimators import read_estimators
from artistools.inputmodel import add_derived_cols_to_modeldata
from artistools.inputmodel import get_cell_angle
from artistools.inputmodel import get_dfmodel_dimensions
from artistools.inputmodel import get_mean_cell_properties_of_angle_bin
from artistools.inputmodel import get_mgi_of_velocity_kms
from artistools.inputmodel import get_modeldata
from artistools.inputmodel import get_modeldata_polars
from artistools.inputmodel import get_modeldata_tuple
from artistools.inputmodel import save_initelemabundances
from artistools.inputmodel import save_modeldata
from artistools.misc import anyexist
from artistools.misc import AppendPath
from artistools.misc import average_direction_bins
from artistools.misc import CustomArgHelpFormatter
from artistools.misc import decode_roman_numeral
from artistools.misc import firstexisting
from artistools.misc import flatten_list
from artistools.misc import get_atomic_number
from artistools.misc import get_bflist
from artistools.misc import get_cellsofmpirank
from artistools.misc import get_composition_data
from artistools.misc import get_composition_data_from_outputfile
from artistools.misc import get_costheta_bins
from artistools.misc import get_costhetabin_phibin_labels
from artistools.misc import get_deposition
from artistools.misc import get_dirbin_labels
from artistools.misc import get_elsymbol
from artistools.misc import get_elsymbols_df
from artistools.misc import get_elsymbolslist
from artistools.misc import get_escaped_arrivalrange
from artistools.misc import get_file_metadata
from artistools.misc import get_filterfunc
from artistools.misc import get_grid_mapping
from artistools.misc import get_inputparams
from artistools.misc import get_ion_stage_roman_numeral_df
from artistools.misc import get_ion_tuple
from artistools.misc import get_ionstring
from artistools.misc import get_linelist_dataframe
from artistools.misc import get_linelist_pldf
from artistools.misc import get_model_name
from artistools.misc import get_mpiranklist
from artistools.misc import get_mpirankofcell
from artistools.misc import get_nprocs
from artistools.misc import get_nu_grid
from artistools.misc import get_phi_bins
from artistools.misc import get_runfolders
from artistools.misc import get_syn_dir
from artistools.misc import get_time_range
from artistools.misc import get_timestep_of_timedays
from artistools.misc import get_timestep_time
from artistools.misc import get_timestep_times
from artistools.misc import get_viewingdirection_costhetabincount
from artistools.misc import get_viewingdirection_phibincount
from artistools.misc import get_viewingdirectionbincount
from artistools.misc import get_vpkt_config
from artistools.misc import get_vspec_dir_labels
from artistools.misc import get_wid_init_at_tmin
from artistools.misc import get_wid_init_at_tmodel
from artistools.misc import get_z_a_nucname
from artistools.misc import linetuple
from artistools.misc import makelist
from artistools.misc import match_closest_time
from artistools.misc import merge_pdf_files
from artistools.misc import namedtuple
from artistools.misc import parse_range
from artistools.misc import parse_range_list
from artistools.misc import read_linestatfile
from artistools.misc import readnoncommentline
from artistools.misc import roman_numerals
from artistools.misc import set_args_from_dict
from artistools.misc import showtimesteptimes
from artistools.misc import split_dataframe_dirbins
from artistools.misc import stripallsuffixes
from artistools.misc import trim_or_pad
from artistools.misc import vec_len
from artistools.misc import zopen
from artistools.misc import zopenpl
from artistools.plottools import set_mpl_style


def get_path(**kwargs: t.Any) -> None:
    print(get_config("path_artistools_dir"))


set_mpl_style()



"""Artistools - light curve functions."""

from artistools.lightcurve.lightcurve import bolometric_magnitude
from artistools.lightcurve.lightcurve import evaluate_magnitudes
from artistools.lightcurve.lightcurve import generate_band_lightcurve_data
from artistools.lightcurve.lightcurve import get_band_lightcurve
from artistools.lightcurve.lightcurve import get_colour_delta_mag
from artistools.lightcurve.lightcurve import get_filter_data
from artistools.lightcurve.lightcurve import get_from_packets
from artistools.lightcurve.lightcurve import get_phillips_relation_data
from artistools.lightcurve.lightcurve import get_sn_sample_bol
from artistools.lightcurve.lightcurve import get_spectrum_in_filter_range
from artistools.lightcurve.lightcurve import plot_phillips_relation_data
from artistools.lightcurve.lightcurve import read_3d_gammalightcurve
from artistools.lightcurve.lightcurve import read_bol_reflightcurve_data
from artistools.lightcurve.lightcurve import read_hesma_lightcurve
from artistools.lightcurve.lightcurve import read_reflightcurve_band_data
from artistools.lightcurve.lightcurve import readfile
from artistools.lightcurve.plotlightcurve import addargs
from artistools.lightcurve.plotlightcurve import main as plot
from artistools.lightcurve.viewingangleanalysis import calculate_peak_time_mag_deltam15
from artistools.lightcurve.viewingangleanalysis import lightcurve_polyfit
from artistools.lightcurve.viewingangleanalysis import make_peak_colour_viewing_angle_plot
from artistools.lightcurve.viewingangleanalysis import make_plot_test_viewing_angle_fit
from artistools.lightcurve.viewingangleanalysis import make_viewing_angle_risetime_peakmag_delta_m15_scatter_plot
from artistools.lightcurve.viewingangleanalysis import parse_directionbin_args
from artistools.lightcurve.viewingangleanalysis import peakmag_risetime_declinerate_init
from artistools.lightcurve.viewingangleanalysis import plot_viewanglebrightness_at_fixed_time
from artistools.lightcurve.viewingangleanalysis import save_viewing_angle_data_for_plotting
from artistools.lightcurve.viewingangleanalysis import second_band_brightness_at_peak_first_band
from artistools.lightcurve.viewingangleanalysis import set_scatterplot_plot_params
from artistools.lightcurve.viewingangleanalysis import set_scatterplot_plotkwargs
from artistools.lightcurve.viewingangleanalysis import update_plotkwargs_for_viewingangle_colorbar
from artistools.lightcurve.viewingangleanalysis import write_viewing_angle_data



# PYTHON_ARGCOMPLETE_OK


import argparse
import math
import sys
import typing as t
from collections.abc import Iterable
from pathlib import Path

import argcomplete
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import polars as pl
from astropy import constants as const
from extinction import apply
from extinction import ccm89

import artistools as at

color_list = list(plt.get_cmap("tab20")(np.linspace(0, 1.0, 20)))


def plot_deposition_thermalisation(axis, axistherm, modelpath, modelname, plotkwargs, args) -> None:
    # if args.logscalex:
    #     axistherm.set_xscale("log")

    # if args.logscaley:
    #     axistherm.set_yscale("log")
    #     axistherm.set_ylim(bottom=0.1, top=1.0)

    if args.plotthermalisation:
        dfmodel, modelmeta = at.inputmodel.get_modeldata(
            modelpath,
            derived_cols=["mass_g", "vel_r_mid"],
        )

        model_mass_grams = dfmodel.mass_g.sum()
        print(f"  model mass: {model_mass_grams / 1.989e33:.3f} Msun")

    depdata = at.get_deposition(modelpath)

    # color_total = axis._get_lines.get_next_color()

    # axis.plot(depdata['tmid_days'], depdata['eps_erg/s/g'] * model_mass_grams, **dict(
    #     plotkwargs, **{
    #         'label': plotkwargs['label'] + r' $\dot{\epsilon}_{\alpha\beta^\pm\gamma}$',
    #         'linestyle': 'dashed',
    #         'color': color_total,
    #     }))

    # axis.plot(depdata['tmid_days'], depdata['total_dep_Lsun'] * 3.826e33, **dict(
    #     plotkwargs, **{
    #         'label': plotkwargs['label'] + r' $\dot{E}_{dep,\alpha\beta^\pm\gamma}$',
    #         'linestyle': 'dotted',
    #         'color': color_total,
    #     }))
    # if args.plotthermalisation:
    #     # f = depdata['eps_erg/s/g'] / depdata['Qdot_ana_erg/s/g']
    #     f = depdata['total_dep_Lsun'] * 3.826e33 / (depdata['eps_erg/s/g'] * model_mass_grams)
    #     axistherm.plot(depdata['tmid_days'], f, **dict(
    #         plotkwargs, **{
    #             'label': plotkwargs['label'] + r' $\dot{E}_{dep}/\dot{E}_{rad}$',
    #             'linestyle': 'solid',
    #             'color': color_total,
    #         }))

    color_gamma = axis._get_lines.get_next_color()  # noqa: SLF001
    color_gamma = axis._get_lines.get_next_color()  # noqa: SLF001

    # axis.plot(depdata['tmid_days'], depdata['eps_gamma_Lsun'] * 3.826e33, **dict(
    #     plotkwargs, **{
    #         'label': plotkwargs['label'] + r' $\dot{E}_{rad,\gamma}$',
    #         'linestyle': 'dashed',
    #         'color': color_gamma,
    #     }))

    gammadep_lsun = depdata["gammadep_Lsun"]

    axis.plot(
        depdata["tmid_days"],
        gammadep_lsun * 3.826e33,
        **{
            **plotkwargs,
            "label": plotkwargs["label"] + r" $\dot{E}_{dep,\gamma}$",
            "linestyle": "dashed",
            "color": color_gamma,
        },
    )

    color_beta = axis._get_lines.get_next_color()  # noqa: SLF001

    if "eps_elec_Lsun" in depdata:
        axis.plot(
            depdata["tmid_days"],
            depdata["eps_elec_Lsun"] * 3.826e33,
            **{
                **plotkwargs,
                "label": plotkwargs["label"] + r" $\dot{E}_{rad,\beta^-}$",
                "linestyle": "dotted",
                "color": color_beta,
            },
        )

    if "elecdep_Lsun" in depdata:
        axis.plot(
            depdata["tmid_days"],
            depdata["elecdep_Lsun"] * 3.826e33,
            **{
                **plotkwargs,
                "label": plotkwargs["label"] + r" $\dot{E}_{dep,\beta^-}$",
                "linestyle": "dashed",
                "color": color_beta,
            },
        )

    # c23modelpath = Path(
    #     Path.home(), "Google Drive/Shared drives/ARTIS/artis_runs_published/Collinsetal2023/sfho_long_1-35-135Msun"
    # )

    # c23energyrate = at.inputmodel.energyinputfiles.get_energy_rate_fromfile(c23modelpath)
    # c23etot, c23energydistribution_data = at.inputmodel.energyinputfiles.get_etot_fromfile(c23modelpath)

    # dE = np.diff(c23energyrate["rate"] * c23etot)
    # dt = np.diff(c23energyrate["times"] * 24 * 60 * 60)

    # axis.plot(
    #     c23energyrate["times"][1:],
    #     dE / dt * 0.308,
    #     color="grey",
    #     linestyle="--",
    #     zorder=20,
    #     label=r"Collins+23 $\dot{E}_{rad,\beta^-}$",
    # )

    # color_alpha = axis._get_lines.get_next_color()
    color_alpha = "C1"

    plotalphadep = False
    if plotalphadep:
        if "eps_alpha_ana_Lsun" in depdata:
            axis.plot(
                depdata["tmid_days"],
                depdata["eps_alpha_ana_Lsun"] * 3.826e33,
                **{
                    **plotkwargs,
                    "label": plotkwargs["label"] + r" $\dot{E}_{rad,\alpha}$ analytical",
                    "linestyle": "solid",
                    "color": color_alpha,
                },
            )

        if "eps_alpha_Lsun" in depdata:
            axis.plot(
                depdata["tmid_days"],
                depdata["eps_alpha_Lsun"] * 3.826e33,
                **{
                    **plotkwargs,
                    "label": plotkwargs["label"] + r" $\dot{E}_{rad,\alpha}$",
                    "linestyle": "dashed",
                    "color": color_alpha,
                },
            )

        axis.plot(
            depdata["tmid_days"],
            depdata["alphadep_Lsun"] * 3.826e33,
            **{
                **plotkwargs,
                "label": plotkwargs["label"] + r" $\dot{E}_{dep,\alpha}$",
                "linestyle": "dotted",
                "color": color_alpha,
            },
        )

    if args.plotthermalisation:
        f_gamma = depdata["gammadep_Lsun"] / depdata["eps_gamma_Lsun"]
        axistherm.plot(
            depdata["tmid_days"],
            f_gamma,
            **{
                **plotkwargs,
                "label": modelname + r" $\left(\dot{E}_{dep,\gamma} \middle/ \dot{E}_{rad,\gamma}\right)$",
                "linestyle": "solid",
                "color": color_gamma,
            },
        )

        f_beta = depdata["elecdep_Lsun"] / depdata["eps_elec_Lsun"]
        axistherm.plot(
            depdata["tmid_days"],
            f_beta,
            **{
                **plotkwargs,
                "label": modelname + r" $\left(\dot{E}_{dep,\beta^-} \middle/ \dot{E}_{rad,\beta^-}\right)$",
                "linestyle": "solid",
                "color": color_beta,
            },
        )

        f_alpha = depdata["alphadep_Lsun"] / depdata["eps_alpha_Lsun"]

        axistherm.plot(
            depdata["tmid_days"],
            f_alpha,
            **{
                **plotkwargs,
                "label": modelname + r" $\left(\dot{E}_{dep,\alpha} \middle/ \dot{E}_{rad,\alpha}\right)$",
                "linestyle": "solid",
                "color": color_alpha,
            },
        )

        ejecta_ke: float
        if "vel_r_mid" in dfmodel.columns:
            # vel_r_mid is in cm/s
            ejecta_ke = (0.5 * (dfmodel["mass_g"] / 1000.0) * (dfmodel["vel_r_mid"] / 100.0) ** 2).sum()
        else:
            # vel_r_min_kmps is in km/s
            ejecta_ke = (0.5 * (dfmodel["mass_g"] / 1000.0) * (1000.0 * dfmodel["vel_r_max_kmps"]) ** 2).sum()

        print(f"  ejecta kinetic energy: {ejecta_ke:.2e} [J] = {ejecta_ke * 1e7:.2e} [erg]")

        # velocity derived from ejecta kinetic energy to match Barnes et al. (2016) Section 2.1
        ejecta_v = np.sqrt(2 * ejecta_ke / (model_mass_grams * 1e-3))
        v2 = ejecta_v / (0.2 * 299792458)
        print(f"  Barnes average ejecta velocity: {ejecta_v / 299792458:.2f}c")
        m5 = model_mass_grams / (5e-3 * 1.989e33)  # M / (5e-3 Msun)

        # Barnes et al (2016) scaling form from equation 17, with fiducial t_ineff_gamma of 1.4 days
        t_ineff_gamma = 1.4 * np.sqrt(m5) / v2
        # Barnes et al (2016) equation 33
        barnes_f_gamma = [1 - math.exp(-((t / t_ineff_gamma) ** -2)) for t in depdata["tmid_days"]]

        axistherm.plot(
            depdata["tmid_days"],
            barnes_f_gamma,
            **{**plotkwargs, "label": r"Barnes+16 $f_\gamma$", "linestyle": "dashed", "color": color_gamma},
        )

        e0_beta_mev = 0.5
        # Barnes et al (2016) equation 20
        t_ineff_beta = 7.4 * (e0_beta_mev / 0.5) ** -0.5 * m5**0.5 * (v2 ** (-3.0 / 2))
        # Barnes et al (2016) equation 32
        barnes_f_beta = [
            math.log(1 + 2 * (t / t_ineff_beta) ** 2) / (2 * (t / t_ineff_beta) ** 2) for t in depdata["tmid_days"]
        ]

        axistherm.plot(
            depdata["tmid_days"],
            barnes_f_beta,
            **{**plotkwargs, "label": r"Barnes+16 $f_\beta$", "linestyle": "dashed", "color": color_beta},
        )

        e0_alpha_mev = 6.0
        # Barnes et al (2016) equation 25 times equation 16 for t_peak
        t_ineff_alpha = 4.3 * 1.8 * (e0_alpha_mev / 6.0) ** -0.5 * m5**0.5 * (v2 ** (-3.0 / 2))
        # Barnes et al (2016) equation 32
        barnes_f_alpha = [
            math.log(1 + 2 * (t / t_ineff_alpha) ** 2) / (2 * (t / t_ineff_alpha) ** 2) for t in depdata["tmid_days"]
        ]

        axistherm.plot(
            depdata["tmid_days"],
            barnes_f_alpha,
            **{**plotkwargs, "label": r"Barnes+16 $f_\alpha$", "linestyle": "dashed", "color": color_alpha},
        )


def plot_artis_lightcurve(
    modelpath: str | Path,
    axis,
    lcindex: int = 0,
    label: str | None = None,
    escape_type: t.Literal["TYPE_RPKT", "TYPE_GAMMA"] = "TYPE_RPKT",
    frompackets: bool = False,
    maxpacketfiles: int | None = None,
    axistherm=None,
    directionbins: t.Sequence[int] | None = None,
    average_over_phi: bool = False,
    average_over_theta: bool = False,
    usedegrees: bool = False,
    args=None,
) -> dict[int, pl.DataFrame] | None:
    lcfilename = None
    modelpath = Path(modelpath)
    if Path(modelpath).is_file():  # handle e.g. modelpath = 'modelpath/light_curve.out'
        lcfilename = Path(modelpath).parts[-1]
        modelpath = Path(modelpath).parent

    if not modelpath.is_dir():
        print(f"\nWARNING: Skipping because {modelpath} does not exist\n")
        return None

    modelname = at.get_model_name(modelpath) if label is None else label
    if lcfilename is not None:
        modelname += f" {lcfilename}"
    if not modelname:
        print("====> (no series label)")
    else:
        print(f"====> {modelname}")
    print(f" modelpath: {modelpath.resolve().parts[-1]}")

    if args is not None and args.title:
        axis.set_title(modelname)

    if directionbins is None:
        directionbins = [-1]

    if frompackets:
        lcdataframes = at.lightcurve.get_from_packets(
            modelpath,
            escape_type=escape_type,
            maxpacketfiles=maxpacketfiles,
            directionbins=directionbins,
            average_over_phi=average_over_phi,
            average_over_theta=average_over_theta,
            get_cmf_column=args.plotcmf,
            directionbins_are_vpkt_observers=args.plotvspecpol is not None,
        )
    else:
        if lcfilename is None:
            lcfilename = (
                "light_curve_res.out"
                if directionbins != [-1]
                else "gamma_light_curve.out"
                if escape_type == "TYPE_GAMMA"
                else "light_curve.out"
            )

        try:
            lcpath = at.firstexisting(lcfilename, folder=modelpath, tryzipped=True)
        except FileNotFoundError:
            print(f"WARNING: Skipping {modelname} because {lcfilename} does not exist")
            return None

        lcdataframes = at.lightcurve.readfile(lcpath)

        if average_over_phi:
            lcdataframes = at.average_direction_bins(lcdataframes, overangle="phi")

        if average_over_theta:
            lcdataframes = at.average_direction_bins(lcdataframes, overangle="theta")

    plotkwargs: dict[str, t.Any] = {
        "label": modelname,
        "linestyle": args.linestyle[lcindex],
        "color": args.color[lcindex],
    }
    if args.dashes[lcindex]:
        plotkwargs["dashes"] = args.dashes[lcindex]
    if args.linewidth[lcindex]:
        plotkwargs["linewidth"] = args.linewidth[lcindex]

    # check if doing viewing angle stuff, and if so define which data to use
    dirbins, angle_definition = at.lightcurve.parse_directionbin_args(modelpath, args)

    if args.colorbarcostheta or args.colorbarphi:
        costheta_viewing_angle_bins, phi_viewing_angle_bins = at.get_costhetabin_phibin_labels(usedegrees=usedegrees)
        scaledmap = make_colorbar_viewingangles_colormap()

    lctimemin, lctimemax = (
        float(lcdataframes[dirbins[0]]["time"].to_numpy().min()),
        float(lcdataframes[dirbins[0]]["time"].to_numpy().max()),
    )

    print(f" range of light curve: {lctimemin:.2f} to {lctimemax:.2f} days")
    try:
        nts_last, validrange_start_days, validrange_end_days = at.get_escaped_arrivalrange(modelpath)
        if validrange_start_days is not None and validrange_end_days is not None:
            str_valid_range = f"{validrange_start_days:.2f} to {validrange_end_days:.2f} days"
        else:
            str_valid_range = f"{validrange_start_days} to {validrange_end_days} days"
        print(f" range of validity (last timestep {nts_last}): {str_valid_range}")
    except FileNotFoundError:
        print(
            " range of validity: could not determine due to missing files "
            "(requires deposition.out, input.txt, model.txt)"
        )
        nts_last, validrange_start_days, validrange_end_days = None, -math.inf, math.inf

    colorindex = None
    for dirbin in dirbins:
        lcdata = lcdataframes[dirbin]

        if dirbin != -1:
            print(f" directionbin {dirbin:4d}  {angle_definition[dirbin]}")

            if args.colorbarcostheta or args.colorbarphi:
                plotkwargs["alpha"] = 0.75
                plotkwargs["label"] = None
                # Update plotkwargs with viewing angle colour
                plotkwargs, colorindex = get_viewinganglecolor_for_colorbar(
                    dirbin,
                    costheta_viewing_angle_bins,
                    phi_viewing_angle_bins,
                    scaledmap,
                    plotkwargs,
                    args,
                )
                if args.average_over_phi_angle:
                    plotkwargs["color"] = "lightgrey"
            else:
                # the first dirbin should use the color argument (which has been removed from the color cycle)
                if dirbin != dirbins[0]:
                    plotkwargs["color"] = None
                plotkwargs["label"] = (
                    f"{modelname}\n{angle_definition[dirbin]}" if modelname else angle_definition[dirbin]
                )

        filterfunc = at.get_filterfunc(args)
        if filterfunc is not None:
            lcdata = lcdata.with_columns(
                pl.from_numpy(filterfunc(lcdata["lum"].to_numpy()), schema=["lum"]).get_column("lum")
            )

        if not args.Lsun or args.magnitude:
            # convert luminosity from Lsun to erg/s
            lcdata = lcdata.with_columns(pl.col("lum") * 3.826e33)
            if "lum_cmf" in lcdata.columns:
                lcdata = lcdata.with_columns(pl.col("lum_cmf") * 3.826e33)

        if args.magnitude:
            # convert to bol magnitude
            lcdata["mag"] = 4.74 - (2.5 * np.log10(lcdata["lum"] / const.L_sun.to("erg/s").value))
            ycolumn = "mag"
        else:
            ycolumn = "lum"

        if (
            args.average_over_phi_angle
            and dirbin % at.get_viewingdirection_costhetabincount() == 0
            and (args.colorbarcostheta or args.colorbarphi)
        ):
            plotkwargs["color"] = scaledmap.to_rgba(colorindex)  # Update colours for light curves averaged over phi
            plotkwargs["zorder"] = 10

        # show the parts of the light curve that are outside the valid arrival range as partially transparent
        if validrange_start_days is None or validrange_end_days is None:
            # entire range is invalid
            lcdata_before_valid = lcdata
            lcdata_after_valid = pl.from_pandas(pd.DataFrame(data=None, columns=lcdata.columns))
            lcdata_valid = pl.from_pandas(pd.DataFrame(data=None, columns=lcdata.columns))
        else:
            lcdata_valid = lcdata.filter(pl.col("time").is_between(validrange_start_days, validrange_end_days))
            if lcdata_valid.is_empty():
                # valid range doesn't contain any data points
                lcdata_before_valid = lcdata
                lcdata_after_valid = pl.from_pandas(pd.DataFrame(data=None, columns=lcdata.columns))
            else:
                lcdata_before_valid = lcdata.filter(pl.col("time") <= lcdata_valid["time"].min())
                lcdata_after_valid = lcdata.filter(pl.col("time") >= lcdata_valid["time"].max())

        if args.plotinvalidpart:
            plotkwargs_invalidrange = plotkwargs.copy()
            plotkwargs_invalidrange.update({"label": None, "alpha": 0.5})
            axis.plot(lcdata_before_valid["time"], lcdata_before_valid[ycolumn], **plotkwargs_invalidrange)
            axis.plot(lcdata_after_valid["time"], lcdata_after_valid[ycolumn], **plotkwargs_invalidrange)
        elif lcdata_valid.is_empty():
            print("  WARNING: No data points in valid range")

        axis.plot(lcdata_valid["time"], lcdata_valid[ycolumn], **plotkwargs)
        if args.print_data:
            print(lcdata[["time", ycolumn, "lum_cmf"]])

        if args.plotcmf:
            plotkwargs["linewidth"] = 1
            plotkwargs["label"] += " (cmf)"
            plotkwargs["linestyle"] = "dashed"
            # plotkwargs['color'] = 'tab:orange'
            axis.plot(lcdata["time"], lcdata["lum_cmf"], **plotkwargs)

    if args.plotdeposition or args.plotthermalisation:
        plot_deposition_thermalisation(axis, axistherm, modelpath, modelname, plotkwargs, args)

    return lcdataframes


def make_lightcurve_plot(
    modelpaths: t.Sequence[str | Path],
    filenameout: str | Path,
    frompackets: bool = False,
    escape_type: t.Literal["TYPE_RPKT", "TYPE_GAMMA"] = "TYPE_RPKT",
    maxpacketfiles: int | None = None,
    args=None,
) -> None:
    """Use light_curve.out or light_curve_res.out files to plot light curve."""
    conffigwidth = float(at.get_config()["figwidth"])
    fig, axis = plt.subplots(
        nrows=1,
        ncols=1,
        sharex=True,
        figsize=(args.figscale * conffigwidth, args.figscale * conffigwidth / 1.6),
        tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0},
    )

    if args.plotthermalisation:
        figtherm, axistherm = plt.subplots(
            nrows=1,
            ncols=1,
            sharex=True,
            figsize=(args.figscale * conffigwidth, args.figscale * conffigwidth / 1.6),
            tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0},
        )
    else:
        axistherm = None

    # take any specified colours our of the cycle
    colors = [
        color for i, color in enumerate(plt.rcParams["axes.prop_cycle"].by_key()["color"]) if f"C{i}" not in args.color
    ]
    axis.set_prop_cycle(color=colors)
    reflightcurveindex = 0

    plottedsomething = False
    lcindex = 0
    for modelpath in modelpaths:
        if not Path(modelpath).is_dir() and not Path(modelpath).exists() and "." in str(modelpath):
            bolreflightcurve = Path(modelpath)

            dflightcurve, metadata = at.lightcurve.read_bol_reflightcurve_data(bolreflightcurve)
            lightcurvelabel = args.label[lcindex] or metadata.get("label", bolreflightcurve)
            color = ["0.0", "0.5", "0.7"][reflightcurveindex]
            plotkwargs = {"label": lightcurvelabel, "color": color, "zorder": 0}
            if (
                "luminosity_errminus_erg/s" in dflightcurve.columns
                and "luminosity_errplus_erg/s" in dflightcurve.columns
            ):
                axis.errorbar(
                    dflightcurve["time_days"],
                    dflightcurve["luminosity_erg/s"],
                    yerr=[dflightcurve["luminosity_errminus_erg/s"], dflightcurve["luminosity_errplus_erg/s"]],
                    fmt="o",
                    capsize=3,
                    **plotkwargs,
                )
            else:
                axis.scatter(dflightcurve["time_days"], dflightcurve["luminosity_erg/s"], **plotkwargs)
            print(f"====> {lightcurvelabel}")
            reflightcurveindex += 1
            plottedsomething = True
        else:
            dirbin = args.plotviewingangle or (args.plotvspecpol or [-1])
            lcdataframes = plot_artis_lightcurve(
                modelpath=modelpath,
                lcindex=lcindex,
                label=args.label[lcindex],
                axis=axis,
                escape_type=escape_type,
                frompackets=frompackets,
                maxpacketfiles=maxpacketfiles,
                axistherm=axistherm,
                directionbins=dirbin,
                average_over_phi=args.average_over_phi_angle,
                average_over_theta=args.average_over_theta_angle,
                usedegrees=args.usedegrees,
                args=args,
            )
            plottedsomething = plottedsomething or (lcdataframes is not None)

        if plottedsomething:
            lcindex += 1

    if args.reflightcurves:
        for bolreflightcurve in args.reflightcurves:
            if args.Lsun:
                print("Check units - trying to plot ref light curve in erg/s")
                sys.exit(1)
            bollightcurve_data, metadata = at.lightcurve.read_bol_reflightcurve_data(bolreflightcurve)
            axis.scatter(
                bollightcurve_data["time_days"],
                bollightcurve_data["luminosity_erg/s"],
                label=metadata.get("label", bolreflightcurve),
                color="k",
            )
            plottedsomething = True

    assert plottedsomething

    if args.magnitude:
        axis.invert_yaxis()

    if args.xmin is not None:
        axis.set_xlim(left=args.xmin)
    if args.xmax is not None:
        axis.set_xlim(right=args.xmax)
    if args.ymin is not None:
        axis.set_ylim(bottom=args.ymin)
    if args.ymax is not None:
        axis.set_ylim(top=args.ymax)
    # axis.set_ylim(bottom=-0.1, top=1.3)

    if not args.nolegend:
        axis.legend(loc="best", handlelength=2, frameon=args.legendframeon, numpoints=1, prop={"size": 9})
        if args.plotthermalisation:
            axistherm.legend(
                loc="upper right", handlelength=2, frameon=args.legendframeon, numpoints=1, prop={"size": 9}
            )

    axis.set_xlabel(r"Time [days]")

    if args.magnitude:
        axis.set_ylabel("Absolute Bolometric Magnitude")
    else:
        str_units = r" [{}$\mathrm{{L}}_\odot$]" if args.Lsun else " [{}erg/s]"
        if args.plotdeposition:
            yvarname = r"$L$ or $\dot{{E}}$"
        elif escape_type == "TYPE_GAMMA":
            yvarname = r"$\mathrm{{L}}_\gamma$"
        elif escape_type == "TYPE_RPKT":
            yvarname = r"$\mathrm{{L}}_{{\mathrm{{UVOIR}}}}$"
        else:
            yvarname = r"$\mathrm{{L}}_{{\mathrm{{" + escape_type.replace("_", r"\_") + r"}}}}$"

        axis.set_ylabel(yvarname + str_units)

        if "{" in axis.get_ylabel() and not args.logscaley:
            axis.yaxis.set_major_formatter(at.plottools.ExponentLabelFormatter(axis.get_ylabel(), decimalplaces=1))

    if args.colorbarcostheta or args.colorbarphi:
        costheta_viewing_angle_bins, phi_viewing_angle_bins = at.get_costhetabin_phibin_labels(
            usedegrees=args.usedegrees
        )
        scaledmap = make_colorbar_viewingangles_colormap()
        make_colorbar_viewingangles(phi_viewing_angle_bins, scaledmap, args, ax=axis)

    if args.logscalex:
        axis.set_xscale("log")

    if args.logscaley:
        axis.set_yscale("log")

    if args.show:
        plt.show()

    fig.savefig(str(filenameout), format="pdf")
    print(f"Saved {filenameout}")

    if args.plotthermalisation:
        # axistherm.set_xscale('log')
        axistherm.set_ylabel("Thermalisation ratio")
        axistherm.set_xlabel(r"Time [days]")
        # axistherm.set_xlim(left=0., args.xmax)
        if args.xmin is not None:
            axistherm.set_xlim(left=args.xmin)
        if args.xmax is not None:
            axistherm.set_xlim(right=args.xmax)
        axistherm.set_ylim(bottom=0.0)
        # axistherm.set_ylim(top=1.05)

        # filenameout2 = "plotthermalisation.pdf"
        filenameout2 = str(filenameout).replace(".pdf", "_thermalisation.pdf")
        figtherm.savefig(filenameout2, format="pdf")
        print(f"Saved {filenameout2}")

    plt.close()


def create_axes(args):
    if "labelfontsize" in args:
        font = {"size": args.labelfontsize}
        mpl.rc("font", **font)

    args.subplots = False  # TODO: set as command line arg

    if (args.filter and len(args.filter) > 1) or args.subplots:
        args.subplots = True
        rows = 2
        cols = 3
    elif args.colour_evolution and len(args.colour_evolution) > 1:
        args.subplots = True
        rows = 1
        cols = 3
    else:
        args.subplots = False
        rows = 1
        cols = 1

    if "figwidth" not in args:
        args.figwidth = at.get_config()["figwidth"] * 1.6 * cols
    if "figheight" not in args:
        args.figheight = at.get_config()["figwidth"] * 1.1 * rows * 1.5

    fig, ax = plt.subplots(
        nrows=rows,
        ncols=cols,
        sharex=True,
        sharey=True,
        figsize=(args.figwidth, args.figheight),
        tight_layout={"pad": 3.0, "w_pad": 0.6, "h_pad": 0.6},
    )  # (6.2 * 3, 9.4 * 3)
    if args.subplots:
        ax = ax.flatten()

    return fig, ax


def get_linelabel(
    modelpath: Path,
    modelname: str,
    modelnumber: int,
    angle: int | None,
    angle_definition: dict[int, str] | None,
    args,
) -> str:
    if angle is not None and angle != -1:
        assert angle_definition is not None
        linelabel = angle_definition[angle] if args.nomodelname else f"{modelname} {angle_definition[angle]}"
        # linelabel = None
        # linelabel = fr"{modelname} $\theta$ = {angle_names[index]}$^\circ$"
        # plt.plot(time, magnitude, label=linelabel, linewidth=3)
    elif args.label:
        linelabel = str(args.label[modelnumber])
    else:
        linelabel = modelname
        # linelabel = 'Angle averaged'

    if linelabel == "None" or linelabel is None:
        linelabel = modelname

    return linelabel


def set_lightcurveplot_legend(ax, args):
    if not args.nolegend:
        if args.subplots:
            ax[args.legendsubplotnumber].legend(
                loc=args.legendposition, frameon=args.legendframeon, fontsize="x-small", ncol=args.ncolslegend
            )
        else:
            ax.legend(
                loc=args.legendposition,
                frameon=args.legendframeon,
                fontsize="small",
                ncol=args.ncolslegend,
                handlelength=0.7,
            )


def set_lightcurve_plot_labels(fig, ax, filternames_conversion_dict, args, band_name=None):
    ylabel = None
    if args.subplots:
        if args.filter:
            ylabel = "Absolute Magnitude"
        if args.colour_evolution:
            ylabel = r"$\Delta$m"
        fig.text(0.5, 0.025, "Time Since Explosion [days]", ha="center", va="center")
        fig.text(0.02, 0.5, ylabel, ha="center", va="center", rotation="vertical")
    else:
        if args.filter and band_name in filternames_conversion_dict:
            ylabel = f"{filternames_conversion_dict[band_name]} Magnitude"
        elif args.filter:
            ylabel = f"{band_name} Magnitude"
        elif args.colour_evolution:
            ylabel = r"$\Delta$m"
        ax.set_ylabel(ylabel, fontsize=args.labelfontsize)  # r'M$_{\mathrm{bol}}$'
        ax.set_xlabel("Time Since Explosion [days]", fontsize=args.labelfontsize)
    if ylabel is None:
        print("failed to set ylabel")
        sys.exit(1)
    return fig, ax


def make_colorbar_viewingangles_colormap():
    norm = mpl.colors.Normalize(vmin=0, vmax=9)
    scaledmap = mpl.cm.ScalarMappable(cmap="tab10", norm=norm)
    scaledmap.set_array([])
    return scaledmap


def get_viewinganglecolor_for_colorbar(
    angle: int, costheta_viewing_angle_bins, phi_viewing_angle_bins, scaledmap, plotkwargs, args
):
    nphibins = at.get_viewingdirection_phibincount()
    if args.colorbarcostheta:
        costheta_index = angle // nphibins
        colorindex = costheta_index
        plotkwargs["color"] = scaledmap.to_rgba(colorindex)
    if args.colorbarphi:
        phi_index = angle % nphibins
        assert nphibins == 10
        reorderphibins = {5: 9, 6: 8, 7: 7, 8: 6, 9: 5}
        print("Reordering phi bins")
        colorindex = reorderphibins.get(colorindex, phi_index)
        plotkwargs["color"] = scaledmap.to_rgba(colorindex)

    return plotkwargs, colorindex


def make_colorbar_viewingangles(phi_viewing_angle_bins, scaledmap, args, fig=None, ax=None):
    if args.colorbarcostheta:
        # ticklabels = costheta_viewing_angle_bins
        ticklabels = [" -1", " -0.8", " -0.6", " -0.4", " -0.2", " 0", " 0.2", " 0.4", " 0.6", " 0.8", " 1"]
        ticklocs = np.linspace(0, 9, num=11)
        label = "cos θ"
    if args.colorbarphi:
        print("reordered phi bins")
        phi_viewing_angle_bins_reordered = [
            "0",
            "π/5",
            "2π/5",
            "3π/5",
            "4π/5",
            "π",
            "6π/5",
            "7π/5",
            "8π/5",
            "9π/5",
            "2π",
        ]
        ticklabels = phi_viewing_angle_bins_reordered
        # ticklabels = phi_viewing_angle_bins
        ticklocs = np.linspace(0, 9, num=11)
        label = "ϕ bin"

    hidecolorbar = False
    if not hidecolorbar:
        if fig:
            # from mpl_toolkits.axes_grid1.inset_locator import inset_axes

            # cax = plt.axes([0.3, 0.97, 0.45, 0.02])  #2nd and 4th move up and down. 1st left and right. 3rd bar width
            cax = plt.axes([0.2, 0.98, 0.65, 0.04])
            cbar = fig.colorbar(scaledmap, cax=cax, orientation="horizontal")
        else:
            cbar = plt.colorbar(scaledmap, ax=ax)
        if label:
            cbar.set_label(label, rotation=0)
        cbar.locator = mpl.ticker.FixedLocator(ticklocs)
        cbar.formatter = mpl.ticker.FixedFormatter(ticklabels)
        cbar.update_ticks()


def make_band_lightcurves_plot(modelpaths, filternames_conversion_dict, outputfolder, args: argparse.Namespace) -> None:
    # angle_names = [0, 45, 90, 180]
    # plt.style.use('dark_background')

    args.labelfontsize = 22  # TODO: make command line arg
    fig, ax = create_axes(args)

    plotkwargs: dict[str, t.Any] = {}

    if args.colorbarcostheta or args.colorbarphi:
        costheta_viewing_angle_bins, phi_viewing_angle_bins = at.get_costhetabin_phibin_labels(
            usedegrees=args.usedegrees
        )
        scaledmap = make_colorbar_viewingangles_colormap()

    first_band_name = None
    for modelnumber, modelpath in enumerate(Path(m) for m in modelpaths):
        # check if doing viewing angle stuff, and if so define which data to use
        angles, angle_definition = at.lightcurve.parse_directionbin_args(modelpath, args)

        for index, angle in enumerate(angles):
            modelname = at.get_model_name(modelpath)
            print(f"Reading spectra: {modelname} (angle {angle})")
            band_lightcurve_data = at.lightcurve.generate_band_lightcurve_data(
                modelpath, args, angle, modelnumber=modelnumber
            )

            if modelnumber == 0 and args.plot_hesma_model:  # TODO: does this work?
                hesma_model = at.lightcurve.read_hesma_lightcurve(args)
                plotkwargs["label"] = str(args.plot_hesma_model).split("_")[:3]

            for plotnumber, band_name in enumerate(band_lightcurve_data):
                if first_band_name is None:
                    first_band_name = band_name
                time, brightness_in_mag = at.lightcurve.get_band_lightcurve(band_lightcurve_data, band_name, args)

                if args.print_data or args.write_data:
                    txtlinesout = [
                        f"# band: {band_name}",
                        f"# model: {modelname}",
                        "# time_days magnitude",
                    ]
                    txtlinesout.extend(f"{t_d} {m}" for t_d, m in zip(time, brightness_in_mag))
                    txtout = "\n".join(txtlinesout)
                if args.write_data:
                    bandoutfile = (
                        Path(f"band_{band_name}_angle_{angle}.txt") if angle != -1 else Path(f"band_{band_name}.txt")
                    )
                    with bandoutfile.open("w") as f:
                        f.write(txtout)
                    print(f"Saved {bandoutfile}")
                if args.print_data:
                    print(txtout)

                plotkwargs["label"] = get_linelabel(modelpath, modelname, modelnumber, angle, angle_definition, args)
                # plotkwargs['label'] = '\n'.join(wrap(linelabel, 40))  # TODO: could be arg? wraps text in label

                filterfunc = at.get_filterfunc(args)
                if filterfunc is not None:
                    brightness_in_mag = filterfunc(brightness_in_mag)

                # This does the same thing as below -- leaving code in case I'm wrong (CC)
                # if args.plotviewingangle and args.plotviewingangles_lightcurves:
                #     global define_colours_list
                #     plt.plot(time, brightness_in_mag, label=modelname, color=define_colours_list[angle], linewidth=3)

                if modelnumber == 0 and args.plot_hesma_model and band_name in hesma_model:  # TODO: see if this works
                    ax.plot(hesma_model.t, hesma_model[band_name], color="black")

                # axarr[plotnumber].axis([0, 60, -16, -19.5])
                text_key = filternames_conversion_dict.get(band_name, band_name)

                if args.subplots:
                    ax[plotnumber].annotate(
                        text_key,
                        xy=(1.0, 1.0),
                        xycoords="axes fraction",
                        textcoords="offset points",
                        xytext=(-30, -30),
                        horizontalalignment="right",
                        verticalalignment="top",
                    )
                # else:
                #     ax.text(args.xmax * 0.75, args.ymax * 0.95, text_key)

                # if not args.calculate_peak_time_mag_deltam15_bool:

                if args.reflightcurves and modelnumber == 0:
                    if len(angles) > 1 and index > 0:
                        print("already plotted reflightcurve")
                    else:
                        define_colours_list = args.refspeccolors
                        markers = args.refspecmarkers
                        for i, reflightcurve in enumerate(args.reflightcurves):
                            plot_lightcurve_from_refdata(
                                band_lightcurve_data.keys(),
                                reflightcurve,
                                define_colours_list[i],
                                markers[i],
                                filternames_conversion_dict,
                                ax,
                                plotnumber,
                            )

                if len(angles) == 1:
                    if args.color:
                        plotkwargs["color"] = args.color[modelnumber]
                    else:
                        plotkwargs["color"] = define_colours_list[modelnumber]

                if args.colorbarcostheta or args.colorbarphi:
                    # Update plotkwargs with viewing angle colour
                    plotkwargs["label"] = None
                    plotkwargs, _ = get_viewinganglecolor_for_colorbar(
                        angle,
                        costheta_viewing_angle_bins,
                        phi_viewing_angle_bins,
                        scaledmap,
                        plotkwargs,
                        args,
                    )

                if args.linestyle:
                    plotkwargs["linestyle"] = args.linestyle[modelnumber]

                # if not (args.test_viewing_angle_fit or args.calculate_peak_time_mag_deltam15_bool):
                curax = ax[plotnumber] if args.subplots else ax
                if args.subplots:
                    if len(angles) > 1 or (args.plotviewingangle and (modelpath / "specpol_res.out").is_file()):
                        ax[plotnumber].plot(time, brightness_in_mag, linewidth=4, **plotkwargs)
                    # I think this was just to have a different line style for viewing angles....
                    else:
                        ax[plotnumber].plot(time, brightness_in_mag, linewidth=4, **plotkwargs)
                        # if key is not 'bol':
                        #     ax[plotnumber].plot(
                        #         cmfgen_mags['time[d]'], cmfgen_mags[key], label='CMFGEN', color='k', linewidth=3)
                else:
                    curax.plot(
                        time, brightness_in_mag, linewidth=3.5, **plotkwargs
                    )  # color=color, linestyle=linestyle)

    at.set_mpl_style()

    ax = at.plottools.set_axis_properties(ax, args)
    fig, ax = set_lightcurve_plot_labels(fig, ax, filternames_conversion_dict, args, band_name=first_band_name)
    set_lightcurveplot_legend(ax, args)

    if args.colorbarcostheta or args.colorbarphi:
        make_colorbar_viewingangles(phi_viewing_angle_bins, scaledmap, args, fig=fig, ax=ax)

    if args.filter and len(band_lightcurve_data) == 1:
        args.outputfile = Path(outputfolder, f"plot{first_band_name}lightcurves.pdf")
    if args.show:
        plt.show()

    ymin, ymax = (ax[0] if args.subplots else ax).get_ylim()
    if ymin < ymax:
        (ax[0] if args.subplots else ax).invert_yaxis()

    plt.savefig(args.outputfile, format="pdf")
    print(f"Saved figure: {args.outputfile}")


# In case this code is needed again...

# if 'redshifttoz' in args and args.redshifttoz[modelnumber] != 0:
#     # print('time before', time)
#     # print('z', args.redshifttoz[modelnumber])
#     time = np.array(time) * (1 + args.redshifttoz[modelnumber])
#     print(f'Correcting for time dilation at redshift {args.redshifttoz[modelnumber]}')
#     # print('time after', time)
#     linestyle = '--'
#     color = 'darkmagenta'
#     linelabel=args.label[1]
# else:
#     linestyle = '-'
#     color='k'
# plt.plot(time, magnitude, label=linelabel, linewidth=3)

# if (args.magnitude or args.plotviewingangles_lightcurves) and not (
#         args.calculate_peakmag_risetime_delta_m15 or args.save_angle_averaged_peakmag_risetime_delta_m15_to_file
#         or args.save_viewing_angle_peakmag_risetime_delta_m15_to_file or args.test_viewing_angle_fit
#         or args.make_viewing_angle_peakmag_risetime_scatter_plot or
#         args.make_viewing_angle_peakmag_delta_m15_scatter_plot):
#     if args.reflightcurves:
#         colours = args.refspeccolors
#         markers = args.refspecmarkers
#         for i, reflightcurve in enumerate(args.reflightcurves):
#             plot_lightcurve_from_refdata(filters_dict.keys(), reflightcurve, colours[i], markers[i],
#                                       filternames_conversion_dict)


def colour_evolution_plot(modelpaths, filternames_conversion_dict, outputfolder, args):
    args.labelfontsize = 24  # TODO: make command line arg
    angle_counter = 0

    fig, ax = create_axes(args)

    plotkwargs = {}

    for modelnumber, modelpath in enumerate(modelpaths):
        modelpath = Path(modelpath)
        modelname = at.get_model_name(modelpath)
        print(f"Reading spectra: {modelname}")

        angles, angle_definition = at.lightcurve.parse_directionbin_args(modelpath, args)

        for index, angle in enumerate(angles):
            for plotnumber, filters in enumerate(args.colour_evolution):
                filter_names = filters.split("-")
                args.filter = filter_names
                band_lightcurve_data = at.lightcurve.generate_band_lightcurve_data(
                    modelpath, args, angle=angle, modelnumber=modelnumber
                )

                plot_times, colour_delta_mag = at.lightcurve.get_colour_delta_mag(band_lightcurve_data, filter_names)

                plotkwargs["label"] = get_linelabel(modelpath, modelname, modelnumber, angle, angle_definition, args)

                filterfunc = at.get_filterfunc(args)
                if filterfunc is not None:
                    colour_delta_mag = filterfunc(colour_delta_mag)

                if args.color and args.plotviewingangle:
                    print(
                        "WARNING: -color argument will not work with viewing angles for colour evolution plots,"
                        "colours are taken from color_list array instead"
                    )
                    # plotkwargs["color"] = color_list[angle_counter]  # index instaed of angle_counter??
                    angle_counter += 1
                elif args.plotviewingangle:
                    plotkwargs["color"] = color_list[angle_counter]
                    angle_counter += 1
                elif args.color:
                    plotkwargs["color"] = args.color[modelnumber]
                if args.linestyle:
                    plotkwargs["linestyle"] = args.linestyle[modelnumber]

                if args.reflightcurves and modelnumber == 0:
                    if len(angles) > 1 and index > 0:
                        print("already plotted reflightcurve")
                    else:
                        for i, reflightcurve in enumerate(args.reflightcurves):
                            plot_color_evolution_from_data(
                                filter_names,
                                reflightcurve,
                                args.refspeccolors[i],
                                args.refspecmarkers[i],
                                filternames_conversion_dict,
                                ax,
                                plotnumber,
                                args,
                            )

                if args.subplots:
                    ax[plotnumber].plot(plot_times, colour_delta_mag, linewidth=4, **plotkwargs)
                else:
                    ax.plot(plot_times, colour_delta_mag, linewidth=3, **plotkwargs)

                curax = ax[plotnumber] if args.subplots else ax
                curax.invert_yaxis()
                curax.annotate(
                    f"{filter_names[0]}-{filter_names[1]}",
                    xy=(1.0, 1.0),
                    xycoords="axes fraction",
                    textcoords="offset points",
                    xytext=(-30, -30),
                    horizontalalignment="right",
                    verticalalignment="top",
                    fontsize="x-large",
                )

            # UNCOMMENT TO ESTIMATE COLOUR AT TIME B MAX
            # def match_closest_time(reftime):
            #     return ("{}".format(min([float(x) for x in plot_times], key=lambda x: abs(x - reftime))))
            #
            # tmax_B = 17.0  # CHANGE TO TIME OF B MAX
            # tmax_B = float(match_closest_time(tmax_B))
            # print(f'{filter_names[0]} - {filter_names[1]} at t_Bmax ({tmax_B}) = '
            #       f'{diff[plot_times.index(tmax_B)]}')

    fig, ax = set_lightcurve_plot_labels(fig, ax, filternames_conversion_dict, args)
    ax = at.plottools.set_axis_properties(ax, args)
    set_lightcurveplot_legend(ax, args)

    args.outputfile = Path(outputfolder, f"plotcolorevolution{filter_names[0]}-{filter_names[1]}.pdf")
    for i in range(2):
        if filter_names[i] in filternames_conversion_dict:
            filter_names[i] = filternames_conversion_dict[filter_names[i]]
    # plt.text(10, args.ymax - 0.5, f'{filter_names[0]}-{filter_names[1]}', fontsize='x-large')

    if args.show:
        plt.show()
    plt.savefig(args.outputfile, format="pdf")


# Just in case it's needed...

# if 'redshifttoz' in args and args.redshifttoz[modelnumber] != 0:
#     plot_times = np.array(plot_times) * (1 + args.redshifttoz[modelnumber])
#     print(f'Correcting for time dilation at redshift {args.redshifttoz[modelnumber]}')
#     linestyle = '--'
#     color='darkmagenta'
#     linelabel = args.label[1]
# else:
#     linestyle = '-'
#     color='k'
#     color='k'


def plot_lightcurve_from_refdata(
    filter_names, lightcurvefilename, color, marker, filternames_conversion_dict, ax, plotnumber
):
    lightcurve_data, metadata = at.lightcurve.read_reflightcurve_band_data(lightcurvefilename)
    linename = metadata["label"] if plotnumber == 0 else None
    filterdir = Path(at.get_config()["path_artistools_dir"], "data/filters/")

    filter_data = {}
    for plotnumber, filter_name in enumerate(filter_names):
        if filter_name == "bol":
            continue
        with Path(filterdir / f"{filter_name}.txt").open() as f:
            lines = f.readlines()
        lambda0 = float(lines[2])

        if filter_name == "bol":
            continue
        if filter_name in filternames_conversion_dict:
            filter_name = filternames_conversion_dict[filter_name]
        filter_data[filter_name] = lightcurve_data.loc[lightcurve_data["band"] == filter_name]
        # plt.plot(limits_x, limits_y, 'v', label=None, color=color)
        # else:

        if "a_v" in metadata or "e_bminusv" in metadata:
            print("Correcting for reddening")

            clightinangstroms = 3e18
            # Convert to flux, deredden, then convert back to magnitudes
            filters = np.array([lambda0] * len(filter_data[filter_name]["magnitude"]), dtype=float)

            filter_data[filter_name]["flux"] = (
                clightinangstroms / (lambda0**2) * 10 ** -((filter_data[filter_name]["magnitude"] + 48.6) / 2.5)
            )  # gs

            filter_data[filter_name]["dered"] = apply(
                ccm89(filters[:], a_v=-metadata["a_v"], r_v=metadata["r_v"]), filter_data[filter_name]["flux"]
            )

            filter_data[filter_name]["magnitude"] = (
                2.5 * np.log10(clightinangstroms / (filter_data[filter_name]["dered"] * lambda0**2)) - 48.6
            )
        else:
            print("WARNING: did not correct for reddening")
        if len(filter_names) > 1:
            ax[plotnumber].plot(
                filter_data[filter_name]["time"],
                filter_data[filter_name]["magnitude"],
                marker,
                label=linename,
                color=color,
            )
        else:
            ax.plot(
                filter_data[filter_name]["time"],
                filter_data[filter_name]["magnitude"],
                marker,
                label=linename,
                color=color,
                linewidth=4,
            )

        # if linename == 'SN 2018byg':
        #     x_values = []
        #     y_values = []
        #     limits_x = []
        #     limits_y = []
        #     for index, row in filter_data[filter_name].iterrows():
        #         if row['date'] == 58252:
        #             plt.plot(row['time'], row['magnitude'], '*', label=linename, color=color)
        #         elif row['e_magnitude'] != -1:
        #             x_values.append(row['time'])
        #             y_values.append(row['magnitude'])
        #         else:
        #             limits_x.append(row['time'])
        #             limits_y.append(row['magnitude'])
        #     print(x_values, y_values)
        #     plt.plot(x_values, y_values, 'o', label=linename, color=color)
        #     plt.plot(limits_x, limits_y, 's', label=linename, color=color)
    return linename


def plot_color_evolution_from_data(
    filter_names, lightcurvefilename, color, marker, filternames_conversion_dict, ax, plotnumber, args
):
    lightcurve_from_data, metadata = at.lightcurve.read_reflightcurve_band_data(lightcurvefilename)
    filterdir = Path(at.get_config()["path_artistools_dir"], "data/filters/")

    filter_data = []
    for i, filter_name in enumerate(filter_names):
        with (filterdir / Path(f"{filter_name}.txt")).open() as f:
            lines = f.readlines()
        lambda0 = float(lines[2])

        if filter_name in filternames_conversion_dict:
            filter_name = filternames_conversion_dict[filter_name]
        filter_data.append(lightcurve_from_data.loc[lightcurve_from_data["band"] == filter_name])

        if "a_v" in metadata or "e_bminusv" in metadata:
            print("Correcting for reddening")
            if "r_v" not in metadata:
                metadata["r_v"] = metadata["a_v"] / metadata["e_bminusv"]
            elif "a_v" not in metadata:
                metadata["a_v"] = metadata["e_bminusv"] * metadata["r_v"]

            clightinangstroms = 3e18
            # Convert to flux, deredden, then convert back to magnitudes
            filters = np.array([lambda0] * filter_data[i].shape[0], dtype=float)

            filter_data[i]["flux"] = (
                clightinangstroms / (lambda0**2) * 10 ** -((filter_data[i]["magnitude"] + 48.6) / 2.5)
            )

            filter_data[i]["dered"] = apply(
                ccm89(filters[:], a_v=-metadata["a_v"], r_v=metadata["r_v"]), filter_data[i]["flux"]
            )

            filter_data[i]["magnitude"] = (
                2.5 * np.log10(clightinangstroms / (filter_data[i]["dered"] * lambda0**2)) - 48.6
            )

    # for i in range(2):
    #     # if metadata['label'] == 'SN 2018byg':
    #     #     filter_data[i] = filter_data[i][filter_data[i].e_magnitude != -99.00]
    #     if metadata['label'] in ['SN 2016jhr', 'SN 2018byg']:
    #         filter_data[i]['time'] = filter_data[i]['time'].apply(lambda x: round(float(x)))  # round to nearest day

    merge_dataframes = filter_data[0].merge(filter_data[1], how="inner", on=["time"])
    if args.subplots:
        ax[plotnumber].plot(
            merge_dataframes["time"],
            merge_dataframes["magnitude_x"] - merge_dataframes["magnitude_y"],
            marker,
            label=metadata["label"],
            color=color,
            linewidth=4,
        )
    else:
        ax.plot(
            merge_dataframes["time"],
            merge_dataframes["magnitude_x"] - merge_dataframes["magnitude_y"],
            marker,
            label=metadata["label"],
            color=color,
        )


def addargs(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "modelpath",
        default=[],
        nargs="*",
        action=at.AppendPath,
        help="Path(s) to ARTIS folders with light_curve.out or packets files (may include wildcards such as * and **)",
    )

    parser.add_argument("-label", default=[], nargs="*", help="List of series label overrides")

    parser.add_argument("--nolegend", action="store_true", help="Suppress the legend from the plot")

    parser.add_argument("--title", action="store_true", help="Show title of plot")

    parser.add_argument("-color", default=[f"C{i}" for i in range(10)], nargs="*", help="List of line colors")

    parser.add_argument("-linestyle", default=[], nargs="*", help="List of line styles")

    parser.add_argument("-linewidth", default=[], nargs="*", help="List of line widths")

    parser.add_argument("-dashes", default=[], nargs="*", help="Dashes property of lines")

    parser.add_argument(
        "-figscale", type=float, default=1.6, help="Scale factor for plot area. 1.0 is for single-column"
    )

    parser.add_argument("--frompackets", action="store_true", help="Read packets files instead of light_curve.out")

    parser.add_argument("-maxpacketfiles", type=int, default=None, help="Limit the number of packet files read")

    parser.add_argument("--gamma", action="store_true", help="Make light curve from gamma rays instead of R-packets")

    parser.add_argument("-escape_type", default="TYPE_RPKT", help="Type of escaping packets")

    parser.add_argument("-o", "-outputfile", action="store", dest="outputfile", type=Path, help="Filename for PDF file")

    parser.add_argument(
        "--plotcmf",
        "--plot_cmf",
        "--showcmf",
        "--show_cmf",
        action="store_true",
        help="Plot comoving frame light curve",
    )

    parser.add_argument(
        "--plotinvalidpart",
        action="store_true",
        help="Plot the entire light curve including partially accumulated parts (light travel time effects)",
    )

    parser.add_argument("--plotdeposition", action="store_true", help="Plot model deposition rates")

    parser.add_argument(
        "--plotthermalisation", action="store_true", help="Plot thermalisation rates (in separate plot)"
    )

    parser.add_argument("--magnitude", action="store_true", help="Plot light curves in magnitudes")

    parser.add_argument("--Lsun", action="store_true", help="Plot light curves in units of Lsun")

    parser.add_argument(
        "-filter",
        "-band",
        dest="filter",
        type=str,
        nargs="+",
        help=(
            "Choose filter eg. bol U B V R I. Default B. "
            "WARNING: filter names are not case sensitive eg. sloan-r is not r, it is rs"
        ),
    )

    parser.add_argument("-colour_evolution", nargs="*", help="Plot of colour evolution. Give two filters eg. B-V")

    parser.add_argument("--print_data", action="store_true", help="Print plotted data")

    parser.add_argument("--write_data", action="store_true", help="Save data used to generate the plot in a text file")

    parser.add_argument(
        "-plot_hesma_model",
        action="store",
        type=Path,
        default=False,
        help="Plot hesma model on top of lightcurve plot. Enter model name saved in data/hesma directory",
    )

    parser.add_argument(
        "-plotvspecpol", type=int, nargs="+", help="Plot vspecpol. Expects int for spec number in vspecpol files"
    )

    parser.add_argument(
        "-plotviewingangle",
        type=int,
        nargs="+",
        help=(
            "Plot viewing angles. Expects int for angle number in specpol_res.out"
            "use args = -2 to select all the viewing angles"
        ),
    )
    parser.add_argument(
        "--usedegrees",
        action="store_true",
        help="Use degrees instead of radians for viewing angles. Only works with -plotviewingangle",
    )

    parser.add_argument("-ymax", type=float, default=None, help="Plot range: y-axis")

    parser.add_argument("-ymin", type=float, default=None, help="Plot range: y-axis")

    parser.add_argument("-xmax", type=float, default=None, help="Plot range: x-axis")

    parser.add_argument("-xmin", type=float, default=None, help="Plot range: x-axis")

    parser.add_argument("-timemax", type=float, default=None, help="Time max to plot")

    parser.add_argument("-timemin", type=float, default=None, help="Time min to plot")

    parser.add_argument("--logscalex", action="store_true", help="Use log scale for horizontal axis")

    parser.add_argument("--logscaley", action="store_true", help="Use log scale for vertial axis")

    parser.add_argument(
        "-reflightcurves",
        type=str,
        nargs="+",
        dest="reflightcurves",
        help="Also plot reference lightcurves from these files",
    )

    parser.add_argument(
        "-refspeccolors", default=["0.0", "0.3", "0.5"], nargs="*", help="Set a list of color for reference spectra"
    )

    parser.add_argument(
        "-refspecmarkers", default=["o", "s", "h"], nargs="*", help="Set a list of markers for reference spectra"
    )

    parser.add_argument(
        "-filtersavgol",
        nargs=2,
        help="Savitzky-Golay filter. Specify the window_length and poly_order.e.g. -filtersavgol 5 3",
    )

    parser.add_argument(
        "-redshifttoz",
        type=float,
        nargs="+",
        help="Redshift to z = x. Expects array length of number modelpaths.If not to be redshifted then = 0.",
    )

    parser.add_argument("--show", action="store_true", default=False, help="Show plot before saving")

    # parser.add_argument('--calculate_peakmag_risetime_delta_m15', action='store_true',
    #                     help='Calculate band risetime, peak mag and delta m15 values for '
    #                     'the models specified using a polynomial fitting method and '
    #                     'print to screen')

    parser.add_argument(
        "--save_angle_averaged_peakmag_risetime_delta_m15_to_file",
        action="store_true",
        help="Save the band risetime, peak mag and delta m15 values for the angle averaged model lightcurves to file",
    )

    parser.add_argument(
        "--save_viewing_angle_peakmag_risetime_delta_m15_to_file",
        action="store_true",
        help=(
            "Save the band risetime, peak mag and delta m15 values for "
            "all viewing angles specified for plotting at a later time "
            "as these values take a long time to calculate for all "
            "viewing angles. Need to run this command first alongside "
            "-plotviewingangle in order to save the data for the "
            "viewing angles you want to use before making the scatter"
            "plots"
        ),
    )

    parser.add_argument(
        "--test_viewing_angle_fit",
        action="store_true",
        help=(
            "Plots the lightcurves for each  viewing angle along with"
            "the polynomial fit for each viewing angle specified"
            "to check the fit is working properly: use alongside"
            "-plotviewingangle "
        ),
    )

    parser.add_argument(
        "--make_viewing_angle_peakmag_risetime_scatter_plot",
        action="store_true",
        help=(
            "Makes scatter plot of band peak mag with risetime with the "
            "angle averaged values being the solid dot and the errors bars"
            "representing the standard deviation of the viewing angle"
            "distribution"
        ),
    )

    parser.add_argument(
        "--make_viewing_angle_peakmag_delta_m15_scatter_plot",
        action="store_true",
        help=(
            "Makes scatter plot of band peak with delta m15 with the angle"
            "averaged values being the solid dot and the error bars representing "
            "the standard deviation of the viewing angle distribution"
        ),
    )

    parser.add_argument(
        "--include_delta_m40",
        action="store_true",
        help="When calculating delta_m15, calculate delta_m40 as well.Only affects the saved viewing angle data.",
    )

    parser.add_argument(
        "--noerrorbars", action="store_true", help="Don't plot error bars on viewing angle scatter plots"
    )

    parser.add_argument(
        "--noangleaveraged", action="store_true", help="Don't plot angle averaged values on viewing angle scatter plots"
    )

    parser.add_argument(
        "--average_over_phi_angle",
        action="store_true",
        help="Average over phi (azimuthal) viewing angles to make direction bins into polar angle bins",
    )

    # for backwards compatibility with above option
    parser.add_argument(
        "--average_every_tenth_viewing_angle",
        action="store_true",
    )

    parser.add_argument(
        "--average_over_theta_angle",
        action="store_true",
        help="Average over theta (polar) viewing angles to make direction bins into azimuthal angle bins",
    )

    parser.add_argument(
        "-calculate_costheta_phi_from_viewing_angle_numbers",
        type=int,
        nargs="+",
        help=(
            "calculate costheta and phi for each viewing angle given the number of the viewing angle"
            "Expects ints for angle number supplied from the argument of plot viewing angle"
            "use args = -1 to select all viewing angles"
            "Note: this method will only work if the number of angle bins (MABINS) = 100"
            "if this is not the case an error will be printed"
        ),
    )

    parser.add_argument(
        "--colorbarcostheta", action="store_true", help="Colour viewing angles by cos theta and show color bar"
    )

    parser.add_argument("--colorbarphi", action="store_true", help="Colour viewing angles by phi and show color bar")

    parser.add_argument(
        "--colouratpeak", action="store_true", help="Make scatter plot of colour at peak for viewing angles"
    )

    parser.add_argument(
        "--brightnessattime",
        action="store_true",
        help="Make scatter plot of light curve brightness at a given time (requires timedays)",
    )

    parser.add_argument("-timedays", "-time", "-t", type=float, help="Time in days to plot")

    parser.add_argument("--nomodelname", action="store_true", help="Model name not added to linename in legend")

    parser.add_argument(
        "-legendsubplotnumber", type=int, default=1, help="Subplot number to place legend in. Default is subplot[1]"
    )

    parser.add_argument("-legendposition", type=str, default="best", help="Position of legend in plot. Default is best")

    parser.add_argument("-ncolslegend", type=int, default=1, help="Number of columns in legend")

    parser.add_argument("--legendframeon", action="store_true", help="Frame on in legend")


def main(args: argparse.Namespace | None = None, argsraw: t.Sequence[str] | None = None, **kwargs: t.Any) -> None:
    """Plot ARTIS light curve."""
    if args is None:
        parser = argparse.ArgumentParser(formatter_class=at.CustomArgHelpFormatter, description=__doc__)
        addargs(parser)
        at.set_args_from_dict(parser, kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args([] if kwargs else argsraw)
        if args.average_every_tenth_viewing_angle:
            print("WARNING: --average_every_tenth_viewing_angle is deprecated. use --average_over_phi_angle instead")
            args.average_over_phi_angle = True

    at.set_mpl_style()

    if not args.modelpath:
        args.modelpath = ["."]

    if not isinstance(args.modelpath, Iterable):
        args.modelpath = [args.modelpath]

    args.modelpath = at.flatten_list(args.modelpath)
    # flatten the list
    modelpaths = []
    for elem in args.modelpath:
        if isinstance(elem, list):
            modelpaths.extend(elem)
        else:
            modelpaths.append(elem)

    args.color, args.label, args.linestyle, args.dashes, args.linewidth = at.trim_or_pad(
        len(args.modelpath), args.color, args.label, args.linestyle, args.dashes, args.linewidth
    )

    if args.gamma:
        args.escape_type = "TYPE_GAMMA"

    if args.filter:
        defaultoutputfile = "plotlightcurves.pdf"
    elif args.colour_evolution:
        defaultoutputfile = "plot_colour_evolution.pdf"
    elif args.escape_type == "TYPE_GAMMA":
        defaultoutputfile = "plotlightcurve_gamma.pdf"
    elif args.escape_type == "TYPE_RPKT":
        defaultoutputfile = "plotlightcurve.pdf"
    else:
        defaultoutputfile = f"plotlightcurve_{args.escape_type}.pdf"

    if not args.outputfile:
        outputfolder = Path()
        args.outputfile = defaultoutputfile
    elif args.outputfile.is_dir():
        outputfolder = Path(args.outputfile)
        args.outputfile = outputfolder / defaultoutputfile
    else:
        outputfolder = Path()

    filternames_conversion_dict = {"rs": "r", "gs": "g", "is": "i", "zs": "z"}

    # determine if this will be a scatter plot or not
    args.calculate_peak_time_mag_deltam15_bool = False
    if (  # args.calculate_peakmag_risetime_delta_m15 or
        args.save_viewing_angle_peakmag_risetime_delta_m15_to_file
        or args.save_angle_averaged_peakmag_risetime_delta_m15_to_file
        or args.make_viewing_angle_peakmag_risetime_scatter_plot
        or args.make_viewing_angle_peakmag_delta_m15_scatter_plot
    ):
        args.calculate_peak_time_mag_deltam15_bool = True
        at.lightcurve.peakmag_risetime_declinerate_init(modelpaths, filternames_conversion_dict, args)
        return

    if args.colouratpeak:  # make scatter plot of colour at peak, eg. B-V at Bmax
        at.lightcurve.make_peak_colour_viewing_angle_plot(args)
        return

    if args.brightnessattime:
        if args.timedays is None:
            print("Specify timedays")
            sys.exit(1)
        if not args.plotviewingangle:
            args.plotviewingangle = [-1]
        if not args.colorbarcostheta and not args.colorbarphi:
            args.colorbarphi = True
        at.lightcurve.plot_viewanglebrightness_at_fixed_time(Path(modelpaths[0]), args)
        return

    if args.filter:
        make_band_lightcurves_plot(modelpaths, filternames_conversion_dict, outputfolder, args)

    elif args.colour_evolution:
        colour_evolution_plot(modelpaths, filternames_conversion_dict, outputfolder, args)
        print(f"Saved figure: {args.outputfile}")
    else:
        make_lightcurve_plot(
            modelpaths=args.modelpath,
            filenameout=args.outputfile,
            frompackets=args.frompackets,
            escape_type=args.escape_type,
            maxpacketfiles=args.maxpacketfiles,
            args=args,
        )


if __name__ == "__main__":
    main()
